# Licensed under the GPLv3 - see LICENSE
import warnings

import numpy as np
import astropy.units as u
from astropy.utils import lazyproperty

from ..base.base import (
    FileBase,
    StreamBase, StreamReaderBase, StreamWriterBase,
    FileOpener, FileInfo)
from .header import GSBHeader
from .payload import GSBPayload
from .frame import GSBFrame
from .file_info import GSBTimeStampInfo, GSBStreamReaderInfo


__all__ = ['GSBTimeStampIO', 'GSBFileReader', 'GSBFileWriter',
           'GSBStreamBase', 'GSBStreamReader', 'GSBStreamWriter',
           'open', 'info']


class GSBTimeStampIO(FileBase):
    """Simple reader/writer for GSB time stamp files.

    Wraps a binary filehandle, providing methods `read_timestamp`,
    `write_timestamp`, and `get_frame_rate`.

    Parameters
    ----------
    fh_raw : filehandle
        Filehandle to the timestamp file, opened in binary mode.
    """

    info = GSBTimeStampInfo()

    def read_timestamp(self):
        """Read a single timestamp.

        Returns
        -------
        frame : `~baseband.gsb.GSBHeader`
            With a ``.time`` property that returns the time encoded.
        """
        return GSBHeader.fromfile(self.fh_raw)

    def write_timestamp(self, header=None, **kwargs):
        """Write a single timestamp.

        Parameters
        ----------
        header : `~baseband.gsb.GSBHeader`, optional
            Header holding time to be written to disk.  Can instead give
            keyword arguments to construct a header.
        **kwargs :
            If ``header`` is not given, these are used to initialize one.
        """
        if header is None:
            header = GSBHeader.fromvalues(**kwargs)
        header.tofile(self.fh_raw)

    def get_frame_rate(self):
        """Determine the number of frames per second.

        The frame rate is inferred from the first two timestamps.

        Returns
        -------
        frame_rate : `~astropy.units.Quantity`
            Frames per second.
        """
        with self.temporary_offset(0):
            timestamp0 = self.read_timestamp()
            timestamp1 = self.read_timestamp()
        return (1. / (timestamp1.time - timestamp0.time)).to(u.Hz)


class GSBFileReader(FileBase):
    """Simple reader for GSB data files.

    Wraps a binary filehandle, providing a `read_payload` method to help
    interpret the data.

    Parameters
    ----------
    payload_nbytes : int
        Number of bytes to read.
    nchan : int, optional
        Number of channels.  Default: 1.
    bps : int, optional
        Bits per elementary sample, i.e. per real or imaginary component
        for complex data.  Default: 4.
    complex_data : bool, optional
        Whether data are complex.  Default: `False`.
    """

    def __init__(self, fh_raw, payload_nbytes, nchan=1, bps=4,
                 complex_data=False):
        self.payload_nbytes = payload_nbytes
        self.nchan = nchan
        self.bps = bps
        self.complex_data = complex_data
        super().__init__(fh_raw)

    def __repr__(self):
        return ("{name}(fh_raw={s.fh_raw}, payload_nbytes={s.payload_nbytes}, "
                "nchan={s.nchan}, bps={s.bps}, complex_data={s.complex_data})"
                .format(name=self.__class__.__name__, s=self))

    def read_payload(self):
        """Read a single block.

        Returns
        -------
        frame : `~baseband.gsb.GSBPayload`
            With a ``.data`` property that returns the data encoded.
        """
        return GSBPayload.fromfile(self.fh_raw,
                                   payload_nbytes=self.payload_nbytes,
                                   sample_shape=(self.nchan,), bps=self.bps,
                                   complex_data=self.complex_data)


class GSBFileWriter(FileBase):
    """Simple writer for GSB data files.

    Adds `write_payload` method to the basic binary file wrapper.
    """

    def write_payload(self, data, bps=4):
        """Write single data block.

        Parameters
        ----------
        data : `~numpy.ndarray` or `~baseband.gsb.GSBPayload`
            If an array, ``bps`` needs to be passed in.
        bps : int, optional
            Bits per elementary sample, to use when encoding the payload.
            Ignored if ``data`` is a GSB payload.  Default: 4.
        """
        if not isinstance(data, GSBPayload):
            data = GSBPayload.fromdata(data, bps=bps)
        return data.tofile(self.fh_raw)


class GSBStreamBase(StreamBase):
    """Base for GSB streams."""

    _sample_shape_maker = GSBPayload._sample_shape_maker

    def __init__(self, fh_ts, fh_raw, header0, sample_rate=None,
                 samples_per_frame=None, payload_nbytes=None, nchan=None,
                 bps=None, complex_data=None, **kwargs):

        self.fh_ts = fh_ts
        rawdump = header0.mode == 'rawdump'
        if isinstance(fh_raw, (tuple, list)):
            assert not rawdump
            for pair in fh_raw:
                assert isinstance(pair, (tuple, list))
                assert len(pair) == len(fh_raw[0])
        elif not rawdump:
            fh_raw = ((fh_raw,),)

        complex_data = (complex_data if complex_data is not None else
                        (False if rawdump else True))
        bps = bps if bps is not None else (4 if rawdump else 8)
        nchan = nchan if nchan is not None else (1 if rawdump else 512)
        bpfs = bps * nchan * (2 if complex_data else 1)
        default_frame_rate = (1e8/6/2**22)*u.Hz
        nfiles = 1 if rawdump else len(fh_raw[0])
        # By default, GSB payloads are always 4 MB (which can combine to
        # frames of 8 MB for phased if two streams are used).
        if payload_nbytes is None:
            if samples_per_frame is None:
                if sample_rate is None:
                    payload_nbytes = 2**22
                else:
                    payload_nbytes = int((sample_rate / default_frame_rate
                                          * bpfs / 8 / nfiles)
                                         .to(u.one).round())
            else:
                payload_nbytes = samples_per_frame * bpfs // (8 * nfiles)

        if samples_per_frame is None:
            samples_per_frame = payload_nbytes * 8 // bpfs * nfiles
        elif samples_per_frame != payload_nbytes*nfiles*8/bpfs:
            raise ValueError('inconsistent samples_per_frame, bps, '
                             'complex_data, and payload_nbytes')

        if sample_rate is None:
            sample_rate = samples_per_frame * default_frame_rate

        sample_shape = (nchan,) if rawdump else (len(fh_raw), nchan)

        super().__init__(
            fh_raw, header0, sample_rate=sample_rate,
            samples_per_frame=samples_per_frame, sample_shape=sample_shape,
            bps=bps, complex_data=complex_data, **kwargs)

        self._payload_nbytes = payload_nbytes

    @property
    def payload_nbytes(self):
        """Number of bytes per payload, divided by the number of raw files."""
        return self._payload_nbytes

    def __getattr__(self, attr):
        """Try to get things on the current open file if it is not on self."""
        if attr in {'readable', 'writable', 'seekable', 'closed', 'name'}:
            fh_raw = (self.fh_raw if self.header0.mode == 'rawdump'
                      else self.fh_raw[0][0])
            try:
                return getattr(fh_raw, attr)
            except AttributeError:  # pragma: no cover
                pass
        #  __getattribute__ to raise appropriate error.
        return self.__getattribute__(attr)

    def _set_index(self, header, index):
        if self.header0.mode == 'phased':
            time_offset = index / self._frame_rate
            # mem_block is a rotating modulo-8 value with no meaning.
            header.update(gps_time=self.header0.gps_time + time_offset,
                          pc_time=self.header0.pc_time + time_offset,
                          seq_nr=self.header0['seq_nr'] + index,
                          mem_block=(self.header0['mem_block'] + index) % 8)
        else:
            super()._set_index(header, index)

    def close(self):
        self.fh_ts.close()
        if self.header0.mode == 'rawdump':
            self.fh_raw.close()
        else:
            for fh_pair in self.fh_raw:
                for fh in fh_pair:
                    fh.close()

    def __repr__(self):
        if isinstance(self.fh_raw, (list, tuple)):
            data_name = tuple(tuple(p.name.split('/')[-1] for p in pol)
                              for pol in self.fh_raw)
        else:
            data_name = self.fh_raw.name
        return ("<{s.__class__.__name__} header={s.fh_ts.name}"
                " offset= {s.offset}\n    data={dn}\n"
                "    sample_rate={s.sample_rate:.5g},"
                " samples_per_frame={s.samples_per_frame},\n"
                "    sample_shape={s.sample_shape}, bps={s.bps},\n"
                "    {sub}start_time={s.start_time.isot}>"
                .format(s=self, dn=data_name, sub=(
                    'subset={0}, '.format(self.subset) if
                    self.subset else '')))


class GSBStreamReader(GSBStreamBase, StreamReaderBase):
    """GSB format reader.

    Allows access to GSB files as a continuous series of samples.  Requires
    both a timestamp and one or more corresponding raw data files.

    Parameters
    ----------
    fh_ts : filehandle
        For reading timestamps.
    fh_raw : filehandle, or nested tuple of filehandles
        Raw binary data filehandle(s).  A single file is needed for rawdump,
        and a tuple for phased.  For a nested tuple, the outer tuple determines
        the number of polarizations, and the inner tuple(s) the number of
        streams per polarization.  E.g., ``((polL1, polL2), (polR1, polR2))``
        for two streams per polarization.  A single tuple is interpreted as
        streams of a single polarization.
    sample_rate : `~astropy.units.Quantity`, optional
        Number of complete samples per second, i.e. the rate at which each
        channel of each polarization is sampled.  If `None`, will be
        inferred assuming the frame rate is exactly 0.25165824 s.
    samples_per_frame : int, optional
        Number of complete samples per frame (possibly combining two files).
        Can give ``payload_nbytes`` instead.
    payload_nbytes : int, optional
        Number of bytes per payload (in each raw file separately).
        If both ``samples_per_frame`` and ``payload_nbytes`` are `None`,
        ``payload_nbytes`` is set to ``2**22`` (4 MiB).
    nchan : int, optional
        Number of channels. Default: 1 for rawdump, 512 for phased.
    bps : int, optional
        Bits per elementary sample, i.e. per real or imaginary component for
        complex data.  Default: 4 for rawdump, 8 for phased.
    complex_data : bool, optional
        Whether data are complex.  Default: `False` for rawdump, `True` for
        phased.
    squeeze : bool, optional
        If `True` (default), remove any dimensions of length unity from decoded
        data.
    subset : indexing object or tuple of objects, optional
        Specific components of the complete sample to decode (after possibly
        squeezing).  If a single indexing object is passed, it selects
        (available) polarizations.  If a tuple is passed, the first selects
        polarizations and the second selects channels.  If the tuple is empty
        (default), all components are read.
    verify : bool, optional
        Whether to do basic checks of frame integrity when reading.  The first
        frame of the stream is always checked.  Default: `True`.
    """
    # TODO: right we are not really compatible with StreamReaderBase,
    # since we need to access multiple files.  Can this be solved with
    # FileWriter/FileReader classes that handle timestamps and multiple blocks,
    # combining these into a frame?
    def __init__(self, fh_ts, fh_raw, sample_rate=None, samples_per_frame=None,
                 payload_nbytes=None, nchan=None, bps=None, complex_data=None,
                 squeeze=True, subset=(), verify=True):
        fh_ts = GSBTimeStampIO(fh_ts)
        header0 = fh_ts.read_timestamp()
        super().__init__(
            fh_ts, fh_raw, header0, sample_rate=sample_rate,
            samples_per_frame=samples_per_frame, payload_nbytes=payload_nbytes,
            nchan=nchan, bps=bps, complex_data=complex_data,
            squeeze=squeeze, subset=subset, verify=verify)
        self.fh_ts.seek(0)
        # Replace fh_raw with GSBFileReader instances.
        fr_kwargs = dict(payload_nbytes=self._payload_nbytes,
                         nchan=self._unsliced_shape.nchan,
                         bps=self.bps, complex_data=self.complex_data)
        if header0.mode == 'rawdump':
            self.fh_raw = GSBFileReader(self.fh_raw, **fr_kwargs)
        else:
            self.fh_raw = [[GSBFileReader(fh, **fr_kwargs)
                            for fh in fh_pair] for fh_pair in self.fh_raw]

    info = GSBStreamReaderInfo()

    @lazyproperty
    def _last_header(self):
        """Last header of the timestamp file."""
        with self.fh_ts.temporary_offset() as fh:
            # Guess based on a fixed header size.  In reality, this
            # may be an overestimate as the headers can grow in size,
            # or an underestimate as the last header may be partial.
            # So, search around to be sure.
            fh_size = fh.seek(0, 2)
            guess = max(fh_size // self.header0.nbytes, 1)
            while self.header0.seek_offset(guess) > fh_size:
                guess -= 1
            while self.header0.seek_offset(guess) < fh_size:
                guess += 1

            # Now see if there is indeed a nice header before.
            fh.seek(self.header0.seek_offset(guess-1))
            last_line = fh.readline()
            last_line_tuple = last_line.split()
            # But realize that sometimes an incomplete header is written.
            try:
                if (len(" ".join(last_line_tuple))
                        < len(" ".join(self.header0.words))):
                    raise EOFError
                last_header = self.header0.__class__(last_line_tuple)
                # Check header time can be parsed.
                last_header.time
            except Exception:
                warnings.warn("The last header entry, '{0}', has an incorect "
                              "length. Using the second-to-last entry instead."
                              .format(last_line))
                fh.seek(self.header0.seek_offset(guess-2))
                last_line_tuple = fh.readline().split()
                last_header = self.header0.__class__(last_line_tuple)
        return last_header

    def readable(self):
        """Whether the file can be read and decoded."""
        return self.info.readable

    def _seek_frame(self, index):
        self.fh_ts.seek(self.header0.seek_offset(index))
        if self.header0.mode == 'rawdump':
            self.fh_raw.seek(index * self._payload_nbytes)
        else:
            for fh_pair in self.fh_raw:
                for fh in fh_pair:
                    fh.seek(index * self._payload_nbytes)

    def _fh_raw_read_frame(self):
        return GSBFrame.fromfile(self.fh_ts, self.fh_raw,
                                 payload_nbytes=self._payload_nbytes,
                                 sample_shape=self._unsliced_shape,
                                 bps=self.bps, complex_data=self.complex_data,
                                 verify=self.verify)


class GSBStreamWriter(GSBStreamBase, StreamWriterBase):
    """GSB format writer.

    Encodes and writes sequences of samples to file.

    Parameters
    ----------
    fh_ts : filehandle
        For writing time stamps to storage.
    fh_raw : filehandle, or nested tuple of filehandles
        For writing raw binary data to storage.  A single file is needed for
        rawdump, and a tuple for phased.  For a nested tuple, the outer
        tuple determines the number of polarizations, and the inner tuple(s)
        the number of streams per polarization.  E.g., ``((polL1, polL2),
        (polR1, polR2))`` for two streams per polarization.  A single tuple is
        interpreted as streams of a single polarization.
    header0 : `~baseband.gsb.GSBHeader`
        Header for the first frame, holding time information, etc.
    sample_rate : `~astropy.units.Quantity`, optional
        Number of complete samples per second, i.e. the rate at which each
        channel of each polarization is sampled.  If not given, will be
        inferred assuming the frame rate is exactly 0.25165824 s.
    samples_per_frame : int, optional
        Number of complete samples per frame (possibly combining two files).
        Can give ``payload_nbytes`` instead.
    payload_nbytes : int, optional
        Number of bytes per payload (in each raw file separately).
        If both ``samples_per_frame`` and ``payload_nbytes`` are `None`,
        ``payload_nbytes`` is set to ``2**22`` (4 MiB).
    nchan : int, optional
        Number of channels. Default: 1 for rawdump, 512 for phased.
    bps : int, optional
        Bits per elementary sample, i.e. per real or imaginary component for
        complex data.  Default: 4 for rawdump, 8 for phased.
    complex_data : bool, optional
        Whether data are complex.  Default: `False` for rawdump, `True` for
        phased.
    squeeze : bool, optional
        If `True` (default), `write` accepts squeezed arrays as input, and
        adds any dimensions of length unity.
    """

    def __init__(self, fh_ts, fh_raw, header0=None, sample_rate=None,
                 samples_per_frame=None, payload_nbytes=None, nchan=None,
                 bps=None, complex_data=None, squeeze=True):
        fh_ts = GSBTimeStampIO(fh_ts)
        super().__init__(
            fh_ts, fh_raw, header0, sample_rate=sample_rate,
            samples_per_frame=samples_per_frame, payload_nbytes=payload_nbytes,
            nchan=nchan, bps=bps, complex_data=complex_data, squeeze=squeeze)
        self._frame = GSBFrame.fromdata(
            np.zeros((self.samples_per_frame,) + self._unsliced_shape,
                     (np.complex64 if self.complex_data else np.float32)),
            header=self.header0.copy(), bps=self.bps)

    def _fh_raw_write_frame(self, frame):
        assert frame.valid
        frame.tofile(self.fh_ts, self.fh_raw)

    def flush(self):
        self.fh_ts.flush()
        try:
            self.fh_raw.flush()
        except AttributeError:
            for fh_pair in self.fh_raw:
                for fh in fh_pair:
                    fh.flush()


class GSBFileOpener(FileOpener):

    non_header_keys = FileOpener.non_header_keys | {'raw'}

    # TODO: think whether the scheme with using FileReader can be made to work.
    def __call__(self, name, mode='rs', **kwargs):
        mode = self.normalize_mode(mode)
        # For binary or timestamp files, the normal opener works fine.
        if mode[1] != 's':
            return super().__call__(name, mode, **kwargs)

        # But for stream mode, we need to open both raw and timestamp.
        fh = self.get_fh(name, mode[0]+'t')
        raw = kwargs.pop('raw', None)
        if raw is None:
            raise TypeError("stream missing required argument 'raw'.")

        stream_mode = kwargs.pop('header_mode',
                                 'phased' if isinstance(raw, (list, tuple))
                                 else 'rawdump')

        if stream_mode == 'rawdump':
            fh_raw = self.get_fh(raw, mode[0]+'b')

        else:
            if not isinstance(raw, (list, tuple)):
                raw = ((raw,),)
            elif not isinstance(raw[0], (list, tuple)):
                raw = (raw,)

            fh_raw = tuple(tuple(self.get_fh(p, mode[0]+'b') for p in pol)
                           for pol in raw)

        if mode == 'ws' and 'header0' not in kwargs:
            kwargs['mode'] = stream_mode
            kwargs['header0'] = self.get_header0(kwargs)

        try:
            return self.classes[mode](fh, fh_raw=fh_raw, **kwargs)
        except Exception:
            if fh is not name:
                fh.close()
            if isinstance(raw, (list, tuple)):
                for pol, polfh in zip(raw, fh_raw):
                    for p, pfh in zip(pol, polfh):
                        if pfh is not p:
                            pfh.close()
            elif fh_raw is not raw:
                fh_raw.close()
            raise


open = GSBFileOpener('GSB', header_class=GSBHeader, classes={
    'rt': GSBTimeStampIO,
    'wt': GSBTimeStampIO,
    'rb': GSBFileReader,
    'wb': GSBFileWriter,
    'rs': GSBStreamReader,
    'ws': GSBStreamWriter}).wrapped(module=__name__, doc="""
Open GSB file(s) for reading or writing.

A GSB data set contains a text header file and one or more raw data files.
When the file is opened as text, one gets a standard filehandle, but with
methods to read/write timestamps.  When it is opened as a binary, one
similarly gets methods to read/write frames.  Opened as a stream, the file
is interpreted as a timestamp file, but raw files need to be given too.
This allows access to the stream(s) as series of samples.

Parameters
----------
name : str
    Filename of timestamp or raw data file.
mode : {'rb', 'wb', 'rt', 'wt', 'rs', or 'ws'}, optional
    Whether to open for reading or writing, and as a regular text or binary
    file (for timestamps and data, respectively) or as a stream.
    Default: 'rs', for reading a stream.
**kwargs
    Additional arguments when opening the file as a stream.

--- For both reading and writing of streams :

raw : str or (tuple of) tuple of str
    Name of files holding payload data.  A single file is needed for
    rawdump, and a tuple for phased.  For a nested tuple, the outer tuple
    determines the number of polarizations, and the inner tuple(s) the
    number of streams per polarization.  E.g.,  ``((polL1, polL2),
    (polR1, polR2))`` for two streams per polarization.  A
    single tuple is interpreted as streams of a single polarization.
sample_rate : `~astropy.units.Quantity`, optional
    Number of complete samples per second, i.e. the rate at which each
    channel of each polarization is sampled.  If `None`, will be
    inferred assuming the frame rate is exactly 251.658240 ms.
samples_per_frame : int, optional
    Number of complete samples per frame.  Can give ``payload_nbytes``
    instead.
payload_nbytes : int, optional
    Number of bytes per payload (in each raw file separately).
    If both ``samples_per_frame`` and ``payload_nbytes`` are `None`,
    ``payload_nbytes`` is set to ``2**22`` (4 MiB).
nchan : int, optional
    Number of channels. Default: 1 for rawdump, 512 for phased.
bps : int, optional
    Bits per elementary sample, i.e. per real or imaginary component for
    complex data.  Default: 4 for rawdump, 8 for phased.
complex_data : bool, optional
    Whether data are complex.  Default: `False` for rawdump, `True` for
    phased.
squeeze : bool, optional
    If `True` (default) and reading, remove any dimensions of length unity
    from decoded data.  If `True` and writing, accept squeezed arrays as
    input, and adds any dimensions of length unity.

--- For reading only :  (see `~baseband.gsb.base.GSBStreamReader`)

subset : indexing object or tuple of objects, optional
    Specific components of the complete sample to decode (after possibly
    squeezing).  If a single indexing object is passed, it selects
    (available) polarizations.  If a tuple is passed, the first selects
    polarizations and the second selects channels.  If the tuple is empty
    (default), all components are read.
verify : bool, optional
    Whether to do basic checks of frame integrity when reading.  The first
    frame of the stream is always checked.  Default: `True`.

--- For writing only : (see `~baseband.gsb.base.GSBStreamWriter`)

header0 : `~baseband.gsb.GSBHeader`
    Header for the first frame, holding time information, etc.  Can instead
    give keyword arguments to construct a header.
**kwargs
    If no header is given, an attempt is made to construct one from these.
    For a standard header, this would include the following.

--- Header keywords : (see :meth:`~baseband.gsb.GSBHeader.fromvalues`)

time : `~astropy.time.Time`
    Start time of the file.
header_mode : 'rawdump' or 'phased', optional
    Used to explicitly set the mode of the GSB stream.  Default: 'rawdump'
    if only a single raw file is present, or 'phased' otherwise.
seq_nr : int, optional
    Frame number, only used for phased (default: 0).

Returns
-------
Filehandle
    :class:`~baseband.gsb.base.GSBTimeStampIO` (timestamp), or
    :class:`~baseband.gsb.base.GSBFileReader` or
    :class:`~baseband.gsb.base.GSBFileWriter` (binary), or
    :class:`~baseband.gsb.base.GSBStreamReader` or
    :class:`~baseband.gsb.base.GSBStreamWriter` (stream)
""")


class GSBFileInfo(FileInfo):
    def get_file_info(self, name, **kwargs):
        info = self._get_info(name, 'rt')
        if self.is_ok(info):
            info.used_kwargs = {}
            if 'raw' in kwargs:
                info.missing.pop('raw')
                info.used_kwargs['raw'] = kwargs['raw']

        return info

    def get_stream_info(self, name, file_info, **kwargs):
        used_kwargs = file_info.used_kwargs
        for key in ('sample_rate', 'payload_nbytes', 'samples_per_frame'):
            if key in kwargs:
                used_kwargs[key] = kwargs[key]

        stream_info = self._get_info(name, mode='rs', **used_kwargs)
        if self.is_ok(stream_info):
            stream_info.used_kwargs = used_kwargs

        return stream_info


info = GSBFileInfo.create(globals())



# Licensed under the GPLv3 - see LICENSE
"""
Definitions for VLBI Mark5B Headers.

Implements a Mark5BHeader class used to store header words, and decode/encode
the information therein.

For the specification, see
https://www.haystack.mit.edu/tech/vlbi/mark5/docs/Mark%205B%20users%20manual.pdf
"""
import numpy as np
import astropy.units as u
from astropy.time import Time

from ..base.header import HeaderParser, VLBIHeaderBase, four_word_struct
from ..base.utils import bcd_decode, bcd_encode, fixedvalue, CRC


__all__ = ['CRC16', 'crc16', 'Mark5BHeader']

CRC16 = 0x18005
"""CRC polynomial used for Mark 5B Headers, as a check on the time code.

x^16 + x^15 + x^2 + 1, i.e., 0x18005.
See page 11 of https://www.haystack.mit.edu/tech/vlbi/mark5/docs/230.3.pdf
(defined there for VLBA headers).

This is also CRC-16-IBM mentioned in
https://en.wikipedia.org/wiki/Cyclic_redundancy_check
"""
crc16 = CRC(CRC16)


class Mark5BHeader(VLBIHeaderBase):
    """Decoder/encoder of a Mark5B Frame Header.

    See page 15 of
    https://www.haystack.mit.edu/tech/vlbi/mark5/docs/Mark%205B%20users%20manual.pdf

    Parameters
    ----------
    words : tuple of int, or None
        Four 32-bit unsigned int header words.  If `None`, set to a tuple of
        zeros for later initialisation.
    kday : int or None
        Explicit thousands of MJD of the observation time (needed to remove
        ambiguity in the Mark 5B time stamp).  Can instead pass an approximate
        ``ref_time``.
    ref_time : `~astropy.time.Time` or None
        Reference time within 500 days of the observation time, used to infer
        the full MJD.  Used only if ``kday`` is not given.
    verify : bool, optional
        Whether to do basic verification of integrity.  Default: `True`.

    Returns
    -------
    header : `Mark5BHeader`
    """

    _header_parser = HeaderParser(
        (('sync_pattern', (0, 0, 32, 0xABADDEED)),
         ('user', (1, 16, 16)),
         ('internal_tvg', (1, 15, 1)),
         ('frame_nr', (1, 0, 15)),
         ('bcd_jday', (2, 20, 12)),
         ('bcd_seconds', (2, 0, 20)),
         ('bcd_fraction', (3, 16, 16)),
         ('crc', (3, 0, 16))))
    _sync_pattern = _header_parser.defaults['sync_pattern']
    _invariants = {'sync_pattern'}
    """Keys of invariant parts in all Mark 5B headers."""
    _stream_invariants = _invariants | {'user'}
    """Keys of invariant parts in a given Mark 5B stream."""

    _struct = four_word_struct

    _properties = ('payload_nbytes', 'frame_nbytes', 'complex_data',
                   'kday', 'jday', 'seconds', 'fraction', 'time')
    """Properties accessible/usable in initialisation."""

    kday = None
    """Thousands of MJD, to complement ``jday`` from header."""

    def __init__(self, words, kday=None, ref_time=None, verify=True):
        if kday is not None:
            self.kday = kday
        super().__init__(words, verify=verify)
        if kday is None and ref_time is not None:
            self.infer_kday(ref_time)

    def verify(self):
        """Verify header integrity."""
        assert len(self.words) == 4
        assert self['sync_pattern'] == self._sync_pattern
        assert self.kday is None or (33000 < self.kday < 400000)
        if self.kday is not None:
            assert self.kday % 1000 == 0, "kday must be thousands of MJD."

    def copy(self, **kwargs):
        return super().copy(kday=self.kday, **kwargs)

    @classmethod
    def fromvalues(cls, *, verify=True, **kwargs):
        """Initialise a header from parsed values.

        Here, the parsed values must be given as keyword arguments, i.e., for
        any ``header = cls(<data>)``, ``cls.fromvalues(**header) == header``.

        However, unlike for the :meth:`Mark5BHeader.fromkeys` class method,
        data can also be set using arguments named after methods, such as
        ``jday`` and ``seconds``.

        Given defaults:

        sync_pattern : 0xABADDEED

        Values set by other keyword arguments (if present):

        bcd_jday : from ``jday`` or ``time``
        bcd_seconds : from ``seconds`` or ``time``
        bcd_fraction : from ``fraction`` or ``time`` (may need ``frame_rate``)
        frame_nr : from ``time`` (may need ``frame_rate``)
        """
        # This method exists only to be able to override the docstring.
        return super().fromvalues(verify=verify, **kwargs)

    def update(self, *, time=None, frame_rate=None, crc=None, verify=True,
               **kwargs):
        """Update the header by setting keywords or properties.

        Here, any keywords matching header keys are applied first, and any
        remaining ones are used to set header properties, in the order set
        by the class (in ``_properties``).

        Parameters
        ----------
        time : `~astropy.time.Time`, optional
            A possible new time.  This is updated last.
        frame_rate : `~astropy.units.Quantity`, optional
            The frame rate to use in calculating the frame number from the
            time.  Needed for times at non-integer seconds.
        crc : int or None, optional
            If `None` (default), recalculate the CRC after updating.
        verify : bool, optional
            If `True` (default), verify integrity after updating.
        **kwargs
            Arguments used to set keywords and properties.
        """
        super().update(verify=False, **kwargs)

        if time is not None:
            self.set_time(time, frame_rate=frame_rate)

        if crc is None:
            # Do not use words 2 & 3 directly, so that this works also if part
            # of a VDIF header, where the time information is in words 7 & 8.
            stream = ((((self['bcd_jday'] << 20)
                        + self['bcd_seconds']) << 16)
                      + self['bcd_fraction'])
            crc = crc16(stream)

        self['crc'] = crc
        if verify:
            self.verify()

    def infer_kday(self, ref_time):
        """Uses a reference time to set a header's ``kday``.

        Parameters
        ----------
        ref_time : `~astropy.time.Time`
            Reference time within 500 days of the observation time.
        """
        self.kday = np.around(ref_time.mjd - self.jday,
                              decimals=-3).astype(int)

    @fixedvalue
    def payload_nbytes(cls):
        """Size of the payload in bytes (always 10000 for Mark5B)."""
        return 10000  # 2500 words

    @fixedvalue
    def frame_nbytes(cls):
        """Size of the frame in bytes (always 10016 for Mark5B)."""
        return cls.nbytes + cls.payload_nbytes

    @fixedvalue
    def complex_data(cls):
        """Whether the data are complex (always False for Mark5B)."""
        return False

    @property
    def jday(self):
        """Last three digits of MJD (decoded from 'bcd_jday')."""
        return bcd_decode(self['bcd_jday'])

    @jday.setter
    def jday(self, jday):
        self['bcd_jday'] = bcd_encode(jday)

    @property
    def seconds(self):
        """Integer seconds on day (decoded from 'bcd_seconds')."""
        return bcd_decode(self['bcd_seconds'])

    @seconds.setter
    def seconds(self, seconds):
        self['bcd_seconds'] = bcd_encode(seconds)

    @property
    def fraction(self):
        """Fractional seconds (decoded from 'bcd_fraction').

        The fraction is stored to 0.1 ms accuracy.  Following mark5access, this
        is "unrounded" to give the exact time of the start of the frame for any
        total bit rate below 512 Mbps.  For rates above this value, it is no
        longer guaranteed that subsequent frames have unique rates.

        Note to the above: since a Mark5B frame contains 80000 bits, the total
        bit rate for which times can be unique would in principle be 800 Mbps.
        However, standard VLBI only uses bit rates that are powers of 2 in MHz.
        """
        ns = bcd_decode(self['bcd_fraction']) * 100000
        # "Unround" the nanoseconds, and turn into fractional seconds.
        return (156250 * ((ns + 156249) // 156250)) / 1e9

    @fraction.setter
    def fraction(self, fraction):
        ns = np.around(fraction * 1.e9)
        # From inspecting sample files, the fraction appears to be truncated,
        # not rounded.
        fraction = (ns / 100000).astype(int)
        self['bcd_fraction'] = bcd_encode(fraction)

    def get_time(self, frame_rate=None):
        """Convert year, BCD time code to Time object.

        Calculate time using `jday`, `seconds`, and `fraction` properties
        (which reflect the bcd-encoded 'bcd_jday', 'bcd_seconds' and
        'bcd_fraction' header items), plus `kday` from the initialisation.  See
        https://www.haystack.mit.edu/tech/vlbi/mark5/docs/Mark%205B%20users%20manual.pdf

        Note that some non-compliant files do not have 'bcd_fraction' set.
        For those, the time can still be calculated using the header's
        'frame_nr' by passing in a frame rate.

        Furthermore, fractional seconds are stored only to 0.1 ms accuracy.
        In the code, this is "unrounded" to give the exact time of the start
        of the frame for any total bit rate below 512 Mbps.  For higher rates,
        it is no longer guaranteed that subsequent frames have unique
        `fraction`, and one should pass in an explicit frame rate instead.

        Parameters
        ----------
        frame_rate : `~astropy.units.Quantity`, optional
            Used to calculate the fractional second from the frame number
            instead of from the header's `fraction`.

        Returns
        -------
        `~astropy.time.Time`
        """
        frame_nr = self['frame_nr']
        if frame_nr == 0:
            fraction = 0.
        elif frame_rate is None:
            fraction = self.fraction
            if fraction == 0.:
                raise ValueError('header does not provide correct fractional '
                                 'second (it is zero for non-zero frame '
                                 'number). Please pass in a frame_rate.')
        else:
            fraction = (frame_nr / frame_rate).to_value(u.s)

        return Time(self.kday + self.jday, (self.seconds + fraction) / 86400,
                    format='mjd', scale='utc', precision=9)

    def set_time(self, time, frame_rate=None):
        """
        Convert Time object to BCD timestamp elements and 'frame_nr'.

        For non-integer seconds, the frame number will be calculated if not
        given explicitly. Doing so requires the frame rate.

        Parameters
        ----------
        time : `~astropy.time.Time`
            The time to use for this header.
        frame_rate : `~astropy.units.Quantity`, optional
            For calculating 'frame_nr' from the fractional seconds.
        """
        self.kday = int(time.mjd // 1000) * 1000
        self.jday = int(time.mjd - self.kday)
        seconds = time - Time(self.kday + self.jday, format='mjd')
        int_sec = int(seconds.sec)
        fraction = seconds - int_sec * u.s

        # Round to nearest ns to handle timestamp difference errors.
        if abs(fraction) < 1. * u.ns:
            frame_nr = 0
            frac_sec = 0.
        elif abs(1. * u.s - fraction) < 1. * u.ns:
            int_sec += 1
            frame_nr = 0
            frac_sec = 0.
        else:
            if frame_rate is None:
                raise ValueError("cannot calculate frame rate. Pass it "
                                 "in explicitly.")
            frame_nr = int((fraction * frame_rate).to(u.one).round())
            fraction = frame_nr / frame_rate
            if abs(fraction - 1. * u.s) < 1. * u.ns:
                int_sec += 1
                frame_nr = 0
                frac_sec = 0.
            else:
                frac_sec = fraction.to(u.s).value

        self.seconds = int_sec
        self.fraction = frac_sec
        self['frame_nr'] = frame_nr

    time = property(get_time, set_time)



# Licensed under the GPLv3 - see LICENSE
"""Provide a base class for "info" properties.

Loosely based on `~astropy.utils.data_info.DataInfo`.
"""
import copy
import operator
import warnings

from astropy import units as u
from astropy.time import Time


__all__ = ['info_item', 'InfoBase',
           'FileReaderInfo', 'StreamReaderInfo', 'NoInfo']


class info_item:
    """Like a lazy property, evaluated only once.

    Can be used as a decorator.

    It replaces itself with the evaluation of the function, i.e.,
    it is not a data descriptor.

    Any errors encountered during the evaluation are stored in the
    instances ``errors`` dict.

    Parameters
    ----------
    attr : str or callable, optional
        If a string, assumes we will get that attribute from ``needs``.
        If a callable, it will be called with the instance as its
        argument to calculate the value (i.e., it will behave like a
        property). If ``attr`` is not given, its is set after the fact
        by applying the instance to a function (i.e., using it as a
        decorator), or by defining it as an attribute of a class.
    needs : str or tuple of str
        The attributes that need to be present to get or calculate
        ``attr``.  If ``attr`` is a string, this should be where the
        attribute should be gotten from (e.g., 'header0' or '_parent');
        if not given, the attribute will simply be set to ``default``.
    default : value, optional
        The value to return if the needs are not met.  Default: `None`.
    doc : str, optional
        Docstring of the descriptor.  If not given will be taken from
        ``attr`` if a function, otherwise constructed.  (Hard to access
        from within python, but useful for sphinx documentation.)
    missing : str, optional
        If the value could be calculated or retrieved, but is `None`,
        then add this string to the ``missing`` attribute of the instance.
        Used, e.g., for Mark 5B to give a helpful message if ``bps``
        is not found on the file reader instance.
    copy : bool
        Whether the copy the value if it is retrieved.  This can be
        useful, e.g., if the value is expected to be a `dict` and an
        independent copy should be made.
    """
    _fget = None

    def __init__(self, attr=None, *, needs=(), default=None, doc=None,
                 missing=None, copy=False):
        needs = tuple(needs) if isinstance(needs, (tuple, list)) else (needs,)
        self.needs = needs
        self.default = default
        self.missing = missing
        self.copy = copy
        # attr will be a callable here if item_info is used as a decorator
        # without any arguments. For backwards compatibility, it can also
        # be the name of the attribute although that will more typically
        # pass through __set_name__.  It can still be used to let the name
        # of the info_item be different from the attribute that is gotten;
        # e.g., start_time = info_item('time', needs='header0').
        self._init_wrapup(attr, doc)

    def _init_wrapup(self, attr, doc=None):
        # Finish initialization, or update from __set_name__ or __call__
        if callable(attr):
            self._fget = attr
            self.name = attr.__name__
            doc = attr.__doc__
        elif attr is not None:
            self.name = attr
            if self._fget is None and self.needs:
                full_attr = '.'.join(self.needs+(attr,))
                self._fget = operator.attrgetter(full_attr)
                doc = "Link to " + full_attr.replace('_parent', 'parent')
        if doc and self.__doc__ is self.__class__.__doc__:
            self.__doc__ = doc

    def __set_name__(self, owner, name):
        # This call-back is entered during class definition time,
        # when the instance has been assigned to a class attribute.
        # This will define the name under which the result gets stored.
        # Will normally be the wrapped function or attribute name.
        self._init_wrapup(name)

    def __call__(self, func):
        """For use as a decorator when not yet fully initialized."""
        # We get here if info_item is used as a decorator but with other
        # arguments defined, e.g., as in @info_item(needs='header0')
        if hasattr(self, 'name'):
            raise TypeError(f"assigned {self.__class__.__name__!r}"
                            f"is not callable")
        self._init_wrapup(func)
        return self

    def __get__(self, instance, cls=None):
        if instance is None:
            return self

        if self._fget and all(getattr(instance, need, None) is not None
                              for need in self.needs):
            try:
                value = self._fget(instance)
            except Exception as exc:
                instance.errors[self.name] = exc
                value = self.default
            else:
                if value is None:
                    if self.missing:
                        instance.missing[self.name] = self.missing
                    value = self.default

        else:
            value = self.default

        if self.copy:
            value = copy.copy(value)

        setattr(instance, self.name, value)
        return value

    def __str__(self):
        short_doc = self.__doc__.split('\n')[0]
        return f"{self.name}: {short_doc}"

    def __repr__(self):
        name = self.__class__.__name__
        extra = {a: getattr(self, a) for a in
                 ('needs', 'default', 'missing', 'copy')}
        extra = ', '.join([f"{a}={v}" for a, v in extra.items()
                           if v or a == 'default' and v is not None])
        if extra:
            extra = f"\n{' '*len(name)}  {extra}"
        return f"<{name} {str(self)}{extra}>"


class InfoBase:
    """Container providing a standardized interface to file information.

    In order to ensure that information is always returned, all access
    to the parent should be via `~baseband.base.file_info.info_item`,
    which ensures that any errors are stored in ``self.errors``.
    In addition, it may be useful to capture warnings and store them in
    ``self.warnings``.

    The instance evaluates as `True` if the underlying file is of the right
    format, and can thus, at least in principle, be read (though more
    information may be needed, given in ``missing``, or the file may be
    corrupt futher on).

    Parameters
    ----------
    parent : instance, optional
        Instance of the file or stream reader the ``info`` instance is
        attached too.  `None` if it is the class version.
    """

    attr_names = ()
    """Attributes that the container provides."""

    _parent = None
    closed = info_item(needs='_parent', doc='Whether parent is closed')

    def __init__(self, parent=None):
        if parent is not None:
            self._parent = parent
            if not self.closed:
                for attr in self.attr_names:
                    getattr(self, attr)

    def _up_to_date(self):
        """Determine whether the information we have stored is up to date."""
        if not hasattr(self, '_parent_attrs'):
            # Set it on the class since it cannot change.
            cls = self.__class__
            cls._parent_attrs = tuple(
                attr for attr in dir(cls)
                if not attr.startswith('_')
                and getattr(getattr(cls, attr), 'needs', ()) == ('_parent',))

        return all(getattr(self, attr) == getattr(self._parent, attr, None)
                   for attr in self._parent_attrs)

    def __get__(self, instance, owner_cls):
        if instance is None:
            # Unbound descriptor, nothing to do.
            return self

        # Check if we have a stored and up to date copy.
        info = instance.__dict__.get('info')
        if info is None or not info._up_to_date():
            # If not, create a new instance and fill it.  Notes:
            # - We cannot change "self", as this was created on the class.
            # - We start from scratch rather than determine what is no longer
            #   up to date, since we cannot know what an update may influence
            #   (e.g., for Mark 4, a change in ref_time affect start_time).
            info = instance.__dict__['info'] = self.__class__(parent=instance)

        return info

    def __delete__(self, instance):
        # We need to define either __set__ or __delete__ since this ensures we
        # are treated as a "data descriptor", i.e., that our __get__ will get
        # called even if "info" is present in instance.__dict__; see
        # https://docs.python.org/3/howto/descriptor.html
        # __delete__ is more useful for us.
        instance.__dict__.pop('info', None)

    def __bool__(self):
        return self.format is not None

    def __call__(self):
        """Create a dict with file information.

        This includes information about checks done, possible missing
        information, as well as possible warnings and errors.
        """
        info = {}
        for attr in self.attr_names:
            value = getattr(self, attr)
            if not (value is None or (isinstance(value, dict)
                                      and value == {})):
                info[attr] = value

        return info

    def __repr__(self):
        # Use the repr for display of file information.
        if self._parent is None:
            return '\n'.join(
                [f"{self.__class__.__name__} (unbound) with attributes:"]
                + [f"  {getattr(self.__class__, attr)}"
                   for attr in self.attr_names])

        if self.closed:
            return "File closed. Not parsable."

        result = [self._parent.__class__.__name__.replace('Reader', '')
                  + ' information:']
        for attr in self.attr_names:
            value = getattr(self, attr)
            if isinstance(value, dict):
                prefix = f"\n{attr}: "
                spaces = ' ' * (len(attr)+2)
                if attr == 'missing':
                    for msg in sorted(set(self.missing.values())):
                        keys = sorted(set(key for key in self.missing
                                          if self.missing[key] == msg))
                        result.append(f"{prefix} {', '.join(keys)}: {msg}")
                        prefix = spaces
                else:
                    for key, val in value.items():
                        str_val = str(val) or repr(val)
                        result.append(f"{prefix} {key}: {str_val}")
                        prefix = spaces

            elif value is not None:
                if isinstance(value, Time):
                    value = Time(value, format='isot', precision=9)
                elif attr == 'sample_rate':
                    value = value.to(u.MHz)
                result.append(f"{attr} = {value}")

        if not self:
            result.append('\nNot parsable. Wrong format?')

        return '\n'.join(result)


class FileReaderInfo(InfoBase):
    """Standardized information on file readers.

    The ``info`` descriptor has a number of standard attributes, which are
    determined from arguments passed in opening the file, from the first header
    (``info.header0``) and from possibly scanning the file to determine the
    duration of frames.

    Examples
    --------
    The most common use is simply to print information::

        >>> from baseband.data import SAMPLE_MARK5B
        >>> from baseband import mark5b
        >>> fh = mark5b.open(SAMPLE_MARK5B, 'rb')
        >>> fh.info
        Mark5BFile information:
        format = mark5b
        number_of_frames = 4
        frame_rate = 6400.0 Hz
        bps = 2
        complex_data = False
        readable = False
        <BLANKLINE>
        missing:  nchan: needed to determine sample shape, frame rate, ...
                  kday, ref_time: needed to infer full times.

        >>> fh.close()

        >>> fh = mark5b.open(SAMPLE_MARK5B, 'rb', kday=56000, nchan=8)
        >>> fh.info
        Mark5BFile information:
        format = mark5b
        number_of_frames = 4
        frame_rate = 6400.0 Hz
        sample_rate = 32.0 MHz
        samples_per_frame = 5000
        sample_shape = (8,)
        bps = 2
        complex_data = False
        start_time = 2014-06-13T05:30:01.000000000
        readable = True
        <BLANKLINE>
        checks:  decodable: True
        >>> fh.close()
    """
    attr_names = ('format', 'number_of_frames', 'frame_rate', 'sample_rate',
                  'samples_per_frame', 'sample_shape', 'bps', 'complex_data',
                  'start_time', 'readable',
                  'missing', 'checks', 'errors', 'warnings')
    """Attributes that the container provides."""

    samples_per_frame = info_item(needs='header0', doc=(
        'Number of complete samples in each frame.'))
    sample_shape = info_item(needs='header0', doc=(
        'Shape of each complete sample (e.g., ``(nchan,)``).'))
    bps = info_item(needs='header0', doc=(
        'Number of bits used to encode each elementary sample.'))
    complex_data = info_item(needs='header0', doc=(
        'Whether the data are complex.'))
    start_time = info_item('time', needs='header0', doc=(
        "Time of the first sample."))

    missing = info_item(default={}, copy=True,
                        doc='dict of missing attributes.')
    checks = info_item(default={}, copy=True,
                       doc='dict of checks for readability.')
    errors = info_item(default={}, copy=True,
                       doc='dict of attributes that raised errors.')
    warnings = info_item(default={}, copy=True,
                         doc='dict of attributes that gave warnings.')

    @info_item
    def header0(self):
        """Header of the first frame in the file."""
        # Here, we do not even know whether the file is open or whether we
        # have the right format. We thus use a try/except and filter out all
        # warnings.
        with self._parent.temporary_offset(0) as fh:
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                return fh.read_header()

    @info_item(needs='header0')
    def frame0(self):
        """First frame from the file."""
        # Try reading a frame.  This has no business failing if a
        # frame rate could be determined, but try anyway; maybe file is closed.
        with self._parent.temporary_offset(0) as fh:
            return fh.read_frame()

    @info_item(needs='frame0', default=False)
    def decodable(self):
        """Whether decoding the first frame worked."""
        # Getting the first sample can fail if we don't have the right decoder.
        self.frame0[0]
        return True

    @info_item(needs='header0')
    def format(self):
        """The file format."""
        return self._parent.__class__.__name__.split('File')[0].lower()

    @info_item(needs='header0')
    def frame_rate(self):
        """Number of frames per unit time."""
        return self._parent.get_frame_rate()

    @info_item(needs='header0')
    def number_of_frames(self):
        """Total number of frames."""
        with self._parent.temporary_offset() as fh:
            file_size = fh.seek(0, 2)

        number_of_frames = file_size / self.header0.frame_nbytes
        if number_of_frames % 1 == 0:
            return int(number_of_frames)
        else:
            self.warnings['number_of_frames'] = (
                f"file contains non-integer number "
                f"({number_of_frames}) of frames")
            return None

    @info_item(needs='frame0', default=False)
    def readable(self):
        """Whether the file is readable and decodable."""
        self.checks['decodable'] = self.decodable
        return all(bool(v) for v in self.checks.values())

    @info_item(needs=('frame_rate', 'samples_per_frame'))
    def sample_rate(self):
        """Rate of complete samples per unit time."""
        return self.frame_rate * self.samples_per_frame


class StreamReaderInfo(InfoBase):
    """Standardized information on stream readers.

    The ``info`` descriptor provides a few standard attributes, most of which
    can also be accessed directly on the stream filehandle, and tests basic
    readability of the stream. More detailed information on the underlying
    file is stored in its info, accessible via ``info.file_info`` (and shown
    by ``__repr__``).
    """
    attr_names = ('start_time', 'stop_time', 'sample_rate', 'shape',
                  'format', 'bps', 'complex_data', 'verify', 'readable',
                  'checks', 'errors', 'warnings')
    """Attributes that the container provides."""

    start_time = info_item(needs='_parent', doc=(
        'Time of the first complete sample.'))
    stop_time = info_item(needs='_parent', doc=(
        'Time of the sample just beyond the end of the file.'))
    sample_rate = info_item(needs='_parent', doc=(
        'Complete samples per unit of time.'))
    shape = info_item(needs='_parent', doc=(
        'Equivalent shape of the whole file.'))
    bps = info_item(needs='_parent', doc=(
        'Number of bits used to encode each elementary sample.'))
    complex_data = info_item(needs='_parent', doc=(
        'Whether the data are complex.'))
    verify = info_item(needs='_parent', doc=(
        'The type of verification done by the stream reader.'))

    # Note that we cannot have missing information in streams;
    # they cannot be opened without it.

    checks = info_item(needs='file_info', copy=True, default={},
                       doc='dict of checks for readability.')
    errors = info_item(needs='file_info', copy=True, default={},
                       doc='dict of attributes that raised errors')
    warnings = info_item(needs='file_info', copy=True, default={},
                         doc='dict of attributes that gave warnings')

    @info_item
    def file_info(self):
        """Information from the underlying file reader."""
        # Our regular baseband readers always have info on their raw
        # files, but we should not presume this is the case. E.g., hdf5
        # readers have an h5File underneath, without info.
        return getattr(getattr(self._parent, 'fh_raw', None), 'info', None)

    @info_item
    def format(self):
        """Format of the underlying file."""
        if self.file_info is not None:
            return self.file_info.format
        elif self.continuous is not None:
            return self._parent.__class__.__name__.split('Stream')[0].lower()

    @info_item
    def continuous(self):
        """Check the stream is continuous.

        Tries reading the very end.  If there is a problem, will bisect
        to find the exact offset at which the problem occurs.

        Errors are raised only to the extent verification is done.  Hence,
        if the stream was opened with ``verify=False``, many fewer problems
        will be found, while if it was opened with ``verify='fix'``, then
        for file types that support it, one will get warnings rather than
        exceptions (if the errors are fixable, of course).
        """
        fh = self._parent
        current_offset = fh.tell()
        try:
            with warnings.catch_warnings():
                warnings.simplefilter('error')
                good = -1
                bad = None
                frame = fh._get_index(fh._last_header)
                while frame > good:
                    try:
                        fh.seek(frame * fh.samples_per_frame)
                        fh.read(1)
                    except Exception as exc:
                        # If we know this is the right one, raise,
                        # otherwise start bisection.
                        if frame == good + 1:
                            msg = f"While reading at {fh.tell()}: "
                            if isinstance(exc, UserWarning):
                                self.warnings['continuous'] = msg + str(exc)
                                return 'fixable gaps'
                            else:
                                self.errors['continuous'] = msg + repr(exc)
                                return False
                        bad = frame
                    else:
                        good = frame

                    if bad is not None:
                        # +1 to ensure we get round towards bad if needed.
                        frame = (bad + good + 1) // 2

            return 'no obvious gaps'

        finally:
            fh.seek(current_offset)

    @info_item
    def readable(self):
        """Whether the stream can be read (possibly fixing errors)."""
        if self.file_info is not None and not self.file_info.readable:
            return False

        self.checks['continuous'] = self.continuous
        return all(bool(v) for v in self.checks.values())

    def _up_to_date(self):
        # Beyond open/close, stream readers can only change in how they verify.
        return (self.verify == self._parent.verify
                and self.closed == self._parent.closed)

    def __call__(self):
        """Create a dict with information about the stream and the raw file."""
        info = super().__call__()
        if self.file_info:
            info['file_info'] = self.file_info()
        return info

    def __repr__(self):
        result = super().__repr__()
        if self._parent is None:
            return result

        file_info = getattr(self, 'file_info', None)
        if file_info is not None:
            # Add information from the raw file, but skip atttributes and
            # extra info dicts that have been included already.
            raw_attrs = file_info.attr_names
            try:
                file_info.attr_names = [attr for attr in raw_attrs
                                        if attr not in self.attr_names]
                result += '\n\n' + repr(file_info)
            finally:
                file_info.attr_names = raw_attrs

        return result


class NoInfo:
    """Info class for cases where no useful information was returned.

    Any instance evaluates as `False`, to indicate a file for which
    the information is given is not readable.

    Parameters
    ----------
    info : str
        Information that will be displayed using ``repr``.
    """
    def __init__(self, info=None):
        self.info = info

    def __bool__(self):
        return False

    def __repr__(self):
        return f"No Info: {self.info}"



# Licensed under the GPLv3 - see LICENSE
from importlib import reload

import pytest
import entrypoints

from .. import io as bio, tasks, vdif, base


class TestExistingIOFormat:
    def setup_method(self):
        dir(bio)  # Ensure entries are loaded.
        self.vdif_entry = bio._entries['vdif']

    def teardown_method(self):
        bio._entries['vdif'] = self.vdif_entry
        bio._bad_entries.discard('vdif')
        dir(bio)  # does update.

    def test_io_entry_point(self):
        assert hasattr(bio, 'vdif')
        assert 'vdif' in bio._entries
        assert 'vdif' in bio.FORMATS
        assert isinstance(bio._entries['vdif'], entrypoints.EntryPoint)
        del bio.vdif
        assert 'vdif' not in bio.__dict__
        assert 'vdif' in bio.FORMATS
        assert 'vdif' in dir(bio)
        assert hasattr(bio, 'vdif')
        assert 'vdif' in bio.__dict__
        assert 'vdif' in bio.FORMATS

    def test_fake_bad_vdif(self):
        assert bio.vdif is vdif
        del bio.vdif
        bio._entries['vdif'] = entrypoints.EntryPoint('vdif', 'bad.vdif', '')
        with pytest.raises(AttributeError, match='not loadable'):
            bio.vdif
        assert 'vdif' not in dir(bio)
        assert 'vdif' in bio._bad_entries
        # Does not auto-reload since already known as bad.
        with pytest.raises(AttributeError, match='has no attribute'):
            bio.vdif
        # But will reload if we reload and thus start over.
        reload(bio)
        assert bio.vdif is vdif
        assert 'vdif' in bio.FORMATS


class TestNewIOFormats:
    def setup_entry(self, entry):
        self.added = entry.name
        bio._entries[entry.name] = entry
        bio.FORMATS.append(entry.name)

    def teardown_method(self):
        bio._entries.pop(self.added, None)
        if self.added in bio.FORMATS:
            bio.FORMATS.remove(self.added)
        bio._bad_entries.discard(self.added)
        bio.__dict__.pop(self.added, None)

    def test_find_new(self):
        self.setup_entry(entrypoints.EntryPoint('new', 'baseband.vdif', ''))
        assert 'new' in dir(bio)
        assert 'new' in bio.FORMATS
        assert bio.new is vdif
        # Check that it comes back if we remove it from the module.
        bio.__dict__.pop('new', None)
        assert 'new' not in bio.__dict__
        assert 'new' in bio.FORMATS
        assert 'new' in dir(bio)
        assert bio.new is vdif

    def test_cannot_getattr_bad(self):
        self.setup_entry(entrypoints.EntryPoint('bad', 'really_bad', ''))
        assert 'bad' in dir(bio)
        assert 'bad' in bio.FORMATS
        with pytest.raises(AttributeError, match='not loadable'):
            bio.bad

        assert 'bad' not in dir(bio)
        assert 'bad' not in bio.FORMATS
        with pytest.raises(AttributeError, match='has no attribute'):
            bio.bad

    def test_not_hasattr_bad(self):
        self.setup_entry(entrypoints.EntryPoint('bad', 'really_bad', ''))
        assert 'bad' in dir(bio)
        assert not hasattr(bio, 'bad')
        assert 'bad' not in dir(bio)


class TestTasks:
    def setup_method(self):
        self.tasks_dict = tasks.__dict__.copy()

    def teardown_method(self):
        tasks.__dict__.clear()
        tasks.__dict__.update(self.tasks_dict)

    def test_first(self):
        assert 'vdif_payload_module' not in dir(tasks)

    def test_task_discovery(self, tmpdir, monkeypatch):
        with open(tmpdir.mkdir('task_tests-0.1.dist-info')
                  .join('entry_points.txt'), 'wt') as fw:
            fw.write('[baseband.tasks]\n'
                     'vdif_payload_module = baseband.vdif.payload\n'
                     'vdif_header_all = baseband.vdif.header:__all__\n'
                     '_nomod = baseband.base.utils:__all__\n')
        monkeypatch.syspath_prepend(str(tmpdir))
        # We loaded just the vdif module.
        assert 'vdif_payload_module' in dir(tasks)
        assert 'vdif' not in dir(tasks)
        assert 'payload' not in dir(tasks)
        assert tasks.vdif_payload_module is vdif.payload
        assert 'VDIFPayload' not in dir(tasks)
        # But helpers and everything in it.
        assert 'vdif_header_all' in dir(tasks)
        assert 'header' not in dir(tasks)
        assert 'VDIFHeader' in dir(tasks)
        assert tasks.vdif_header_all is vdif.header
        assert tasks.VDIFHeader is vdif.header.VDIFHeader
        # And what's in utils, but not the name.
        assert '_nomod' not in dir(tasks)
        assert 'CRC' in dir(tasks)
        assert tasks.CRC is base.utils.CRC

    def test_bad_task_definition(self, tmpdir, monkeypatch):
        with open(tmpdir.mkdir('bad_task_tests-0.1.dist-info')
                  .join('entry_points.txt'), 'wt') as fw:
            fw.write('[baseband.tasks]\n'
                     'vdif_payload_module = baseband.vdif.payload\n'
                     'utils = baseband.base.utils.__all__\n'  # typo: . not :
                     'does_not_exist = baseband.does_not_exist\n')

        monkeypatch.syspath_prepend(str(tmpdir))
        assert tasks.vdif_payload_module is vdif.payload
        assert not hasattr(tasks, 'utils')
        assert 'utils' not in dir(tasks)
        assert 'does_not_exist' not in dir(tasks)
        assert len(tasks._bad_entries) == 2
        assert (set(entry.name for entry in tasks._bad_entries)
                == {'utils', 'does_not_exist'})

    @pytest.mark.xfail(entrypoints.get_group_all('baseband.tasks'),
                       reason='cannot test for lack of entry points')
    def test_message_on_empty_tasks(self):
        with pytest.raises(AttributeError, match='No.*entry points found'):
            tasks.does_not_exist

    def test_last(self):
        assert 'vdif_payload_module' not in dir(tasks)



# Licensed under the GPLv3 - see LICENSE
"""Mark 4 VLBI data reader.

Code inspired by Walter Brisken's mark5access.  See
https://github.com/demorest/mark5access.

The format itself is described in detail in
https://www.haystack.mit.edu/tech/vlbi/mark5/docs/230.3.pdf
"""
from .base import open, info  # noqa
from .header import Mark4Header  # noqa
from .payload import Mark4Payload  # noqa
from .frame import Mark4Frame  # noqa



# Licensed under the GPLv3 - see LICENSE
import io
import os
import re
import itertools
from bisect import bisect

import numpy as np
from astropy.utils import lazyproperty


__all__ = ['FileNameSequencer', 'SequentialFileBase',
           'SequentialFileReader', 'SequentialFileWriter',
           'open']


class FileNameSequencer:
    """List-like generator of filenames using a template.

    The template is formatted, filling in any items in curly brackets with
    values from the header.  It is additionally possible to insert a file
    number equal to the indexing value, indicated with '{file_nr}'.

    The length of the instance will be the number of files that exist that
    match the template for increasing values of the file number (when writing,
    it is the number of files that have so far been generated).

    Parameters
    ----------
    template : str
        Template to format to get specific filenames.  Curly bracket item
        keywords are case-sensitive (eg. '{FRAME_NR}' or '{Frame_NR}' will not
        use ``header['frame_nr']``.
    header : dict-like
        Structure holding key'd values that are used to fill in the format.

    Examples
    --------

    >>> from baseband import vdif
    >>> from baseband.helpers import sequentialfile as sf
    >>> vfs = sf.FileNameSequencer('a{file_nr:03d}.vdif')
    >>> vfs[10]
    'a010.vdif'
    >>> from baseband.data import SAMPLE_VDIF
    >>> with vdif.open(SAMPLE_VDIF, 'rb') as fh:
    ...     header = vdif.VDIFHeader.fromfile(fh)
    >>> vfs = sf.FileNameSequencer('obs.edv{edv:d}.{file_nr:05d}.vdif', header)
    >>> vfs[10]
    'obs.edv3.00010.vdif'
    """

    def __init__(self, template, header={}):
        self.items = {}

        def check_and_convert(x):
            string = x.group()
            key = string[1:-1]
            if key != 'file_nr':
                self.items[key] = header[key]
            return string

        self.template = re.sub(r'{\w+[}:]', check_and_convert, template)

    def _process_items(self, file_nr):
        # No check for whether file_nr > len(self), since there may not be a
        # predeterminable length when writing.
        if file_nr < 0:
            file_nr += len(self)
            if file_nr < 0:
                raise IndexError('file number out of range.')

        self.items['file_nr'] = file_nr

    def __getitem__(self, file_nr):
        self._process_items(file_nr)
        return self.template.format(**self.items)

    def __len__(self):
        file_nr = 0
        while os.path.isfile(self[file_nr]):
            file_nr += 1

        return file_nr


class SequentialFileBase:
    """Deal with several files as if they were one contiguous one.

    For details, see `SequentialFileReader` and `SequentialFileWriter`.
    """
    file_nr = None

    def __init__(self, files, mode='rb', opener=None):
        self.files = files
        self.mode = mode
        self.opener = io.open if opener is None else opener
        self._file_sizes = []
        self._file_offsets = [0]
        self._open(0)

    def __getattr__(self, attr):
        """Try to get things on the current open file if it is not on self."""
        if not attr.startswith('_'):
            try:
                return getattr(self.fh, attr)
            except AttributeError:
                pass
        return self.__getattribute__(attr)

    def _open(self, file_nr):
        """Open the ``file_nr``th file of the list of underlying files.

        If a different file was already open, it is closed.  Nothing is done
        if the requested file is already open.
        """
        if file_nr != self.file_nr:
            try:
                fh = self.opener(self.files[file_nr], mode=self.mode)
            except IndexError:
                raise OSError('ran out of files.')
            if self.file_nr is not None:
                self.fh.close()
            self.fh = fh
            self.file_nr = file_nr
            if self.file_nr == len(self._file_sizes):
                file_size = self.file_size
                if file_size is not None:  # can happen for single-file write.
                    self._file_sizes.append(file_size)
                    self._file_offsets.append(self._file_offsets[-1]
                                              + file_size)

    def tell(self):
        """Return the current stream position."""
        return self._file_offsets[self.file_nr] + self.fh.tell()

    def memmap(self, dtype=np.uint8, mode=None, offset=None, shape=None,
               order='C'):
        """Map part of the file in memory.

        Note that the map cannnot span multiple underlying files.
        Parameters are as for `~numpy.memmap`.
        """
        if self.closed:
            raise ValueError('memmap of closed file.')

        dtype = np.dtype(dtype)

        if mode is None:
            mode = self.mode.replace('b', '')

        if offset is not None and offset != self.tell():
            # seek will fail for SequentialFileWriter, so we try to avoid it.
            self.seek(offset)
        elif self.fh.tell() == self._file_sizes[self.file_nr]:
            self._open(self.file_nr + 1)

        if shape is None:
            count = self.size - self.tell()
            if count % dtype.itemsize:
                raise ValueError("size of available data is not a "
                                 "multiple of the data-type size.")
            shape = (count // dtype.itemsize,)
        else:
            if not isinstance(shape, tuple):
                shape = (shape,)
            count = dtype.itemsize
            for k in shape:
                count *= k

        if self.fh.tell() + count > self._file_sizes[self.file_nr]:
            raise ValueError('mmap length exceeds individual file size')

        file_offset = self.fh.tell()
        mm = np.memmap(self.fh, dtype, mode, file_offset, shape, order)
        self.fh.seek(file_offset + count)
        return mm

    def close(self):
        """Close the currently open local file, and therewith the set."""
        if self.file_nr is not None:
            self.fh.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __repr__(self):
        base_repr = ("{0}(files={1}, mode='{2}')"
                     .format(self.__class__.__name__, self.files, self.mode))
        extra = ("At offset: {0}; open file: {1!r}."
                 .format(self.tell(), None if self.file_nr is None else
                         self.files[self.file_nr]))
        return base_repr + "\n# " + extra


class SequentialFileReader(SequentialFileBase):
    """Read several files as if they were one contiguous one.

    Parameters
    ----------
    files : list, tuple, or other iterable of str, filehandle
        The contains the names of the underlying files that should be combined.
        If not a list or tuple, it should allow indexing with positive indices,
        and raise `IndexError` if these are out of range.
    mode : str, optional
        The mode with which the files should be opened (default: 'rb')
    opener : callable, optional
        Function to open a single file (default: `io.open`).
    """

    def __getattr__(self, attr):
        if attr.startswith('read'):
            # Ensure we skip to the next file if needed.
            self.seek(0, 1)

        return super().__getattr__(attr)

    @property
    def file_size(self):
        """Size of the underlying file currently open for reading."""
        offset = self.fh.tell()
        file_size = self.fh.seek(0, 2)
        self.fh.seek(offset)
        return file_size

    @lazyproperty
    def size(self):
        """Size of all underlying files combined."""
        offset = self.tell()
        for i in itertools.count(start=len(self._file_sizes)):
            try:
                self._open(i)
            except Exception:
                break

        self.seek(offset)
        return self._file_offsets[-1]

    def seek(self, offset, whence=0):
        if self.closed:
            raise ValueError('seek of closed file.')

        if whence == 1:
            offset += self.tell()
        elif whence == 2:
            offset += self.size
        elif whence != 0:
            raise ValueError("invalid 'whence'; should be 0, 1, or 2.")

        if offset < 0:
            raise OSError('invalid offset')

        # If the offset is not in the current file, find right one.
        while not (0 <= offset - self._file_offsets[self.file_nr]
                   < self._file_sizes[self.file_nr]):
            # Note that not all files may have been opened at this point.
            # In that case, bisecting would find we're out of the current files
            # and one would open a new one.  The while loop ensures we keep
            # trying until we've got there or reached the end of the files.
            file_nr = bisect(self._file_offsets, offset) - 1
            try:
                self._open(file_nr)
            except OSError:
                # If no files left, put pointer beyond end of last file.
                if file_nr != len(self._file_sizes):  # pragma: no cover
                    raise
                self._open(file_nr - 1)
                break

        self.fh.seek(offset - self._file_offsets[self.file_nr])
        return self.tell()
    seek.__doc__ = io.BufferedIOBase.seek.__doc__

    def read(self, count=None):
        if self.closed:
            raise ValueError('read of closed file.')

        if count is None or count < 0:
            count = max(self.size - self.tell(), 0)

        data = b''
        while count > 0:
            # Go to current offset, possibly opening new file.
            self.seek(0, 1)
            extra = self.fh.read(count)
            if not extra:
                break
            count -= len(extra)
            if not data:  # avoid copies for first read.
                data = extra
            else:
                data += extra

        return data
    read.__doc__ = io.BufferedIOBase.read.__doc__

    def __getstate__(self):
        state = self.__dict__.copy()
        # IOBase instances cannot be pickled, but we can just reopen them
        # when we are unpickled.  Anything else may have internal state that
        # needs preserving (e.g., another SequentialFile), so we will assume
        # it takes care of this itself.
        if isinstance(self.fh, io.IOBase):
            fh = state.pop('fh')
            state['fh_info'] = {
                'file_nr': state.pop('file_nr'),
                'offset': 'closed' if fh.closed else fh.tell()}

        return state

    def __setstate__(self, state):
        fh_info = state.pop('fh_info', None)
        self.__dict__.update(state)
        if fh_info is not None:
            self._open(fh_info['file_nr'])
            if fh_info['offset'] != 'closed':
                self.fh.seek(fh_info['offset'])
            else:
                self.fh.close()


class SequentialFileWriter(SequentialFileBase):
    """Write several files as if they were one contiguous one.

    Note that the file is not seekable and readable.

    Parameters
    ----------
    files : list, tuple, or other iterable of str, filehandle
        The contains the names of the underlying files that should be combined.
        If not a list or tuple, it should allow indexing with positive indices
        (e.g., returning a name as derived from a template).  It should raise
        raise `IndexError` if the index is out of range.
    mode : str, optional
        The mode with which the files should be opened (default: 'w+b'). If
        this does not include '+' for reading, memory maps are not possibe.
    file_size : int, optional
        The maximum size a file is allowed to have.  Default: `None`, which
        means it is unlimited and only a single file will be written (making
        using this class somewhat pointless).
    opener : callable, optional
        Function to open a single file (default: `io.open`).
    """

    def __init__(self, files, mode='w+b', file_size=None, opener=None):
        self.file_size = file_size
        super().__init__(files, mode, opener)

    def write(self, data):
        if self.closed:
            raise ValueError('write to closed file.')
        offset0 = self.tell()
        if self.file_size is not None:
            remaining = self.file_size - self.fh.tell()
            while len(data) > remaining:
                self.fh.write(data[:remaining])
                data = data[remaining:]
                self._open(self.file_nr + 1)
                remaining = self.file_size

        self.fh.write(data)
        return self.tell() - offset0
    write.__doc__ = io.BufferedIOBase.write.__doc__

    def memmap(self, dtype=np.uint8, mode=None, offset=None, shape=None,
               order='C'):
        """Map part of the file in memory.  Cannnot span file boundaries."""
        if shape is None:
            raise ValueError('cannot make writable memmap without shape.')
        return super().memmap(dtype, mode, offset, shape, order)


def open(files, mode='rb', file_size=None, opener=None):
    """Read or write several files as if they were one contiguous one.

    Parameters
    ----------
    files : list, tuple, or other iterable of str, filehandle
        Contains the names of the underlying files that should be combined,
        ordered in time.  If not a list or tuple, it should allow indexing with
        positive indices, and raise `IndexError` if these are out of range.
    mode : str, optional
        The mode with which the files should be opened (default: 'rb').
    file_size : int, optional
        For writing, the maximum size of a file, beyond which a new file should
        be opened.  Default: `None`, which means it is unlimited and only a
        single file will be written.
    opener : callable, optional
        Function to open a single file (default: `io.open`).

    Notes
    -----
    The returned reader/writer will have a ``memmap`` method with which part of
    the files can be mapped to memory (like with `~numpy.memmap`), as long as
    those parts do not span files (and the underlying files are regular ones).
    For writing, this requires opening in read-write mode (i.e., 'w+b').

    Methods other than ``read``, ``write``, ``seek``, ``tell``, and ``close``
    are tried on the underlying file.  This implies, e.g., ``readline`` is
    possible, though the line cannot span multiple files.

    The reader assumes the sequence of files is **contiguous in time**, ie.
    with no gaps in the data.
    """
    if 'r' in mode:
        if file_size is not None:
            raise TypeError("cannot pass in 'file_size' for reading.")
        return SequentialFileReader(files, mode, opener=opener)
    elif 'w' in mode:
        return SequentialFileWriter(files, mode, file_size=file_size,
                                    opener=opener)
    else:
        raise ValueError("invalid mode '{0}'".format(mode))



# Licensed under the GPLv3 - see LICENSE
import pytest
import numpy as np
from astropy import units as u
from astropy.time import Time

from ... import mark5b
from ...data import SAMPLE_MARK5B as SAMPLE_FILE
from ...base.base import HeaderNotFoundError


class TestCorruptSampleCopy:
    def setup_class(cls):
        with open(SAMPLE_FILE, 'rb') as fh:
            cls.sample_bytes = fh.read()
        with mark5b.open(SAMPLE_FILE, 'rs', sample_rate=32*u.MHz,
                         kday=56000, nchan=8, bps=2) as fs:
            cls.frame_nbytes = fs.header0.frame_nbytes
            cls.frame_rate = fs._frame_rate
            cls.start_time = fs.start_time
            cls.stop_time = fs.stop_time
            cls.data = fs.read()
            cls.frame3 = fs._frame

    def expected_bad_frames(self, missing):
        (start_f, start_r), (stop_f, stop_r) = [
            divmod(s, self.frame_nbytes)
            for s in (missing.start, missing.stop-1)]

        if start_r < 5:  # invariants touched?
            start_f -= 1

        return start_f, stop_f+1

    @pytest.mark.parametrize('missing,expected_bad_start,expected_bad_stop', [
        (slice(20032, 20033), 1, 3),  # First byte of header of frame 2.
        (slice(20096, 20100), 2, 3),  # Part of payload of frame 2.
        (slice(12000, 22000), 1, 3),  # Parts of 1-2.
        (slice(30060, 30070), 3, 4),  # Part of header of frame 3.
        (slice(40063, 40064), 3, 4)])  # Last byte of last frame.
    def test_expected_bad_frames(self, missing, expected_bad_start,
                                 expected_bad_stop):
        bad_start, bad_stop = self.expected_bad_frames(missing)
        assert bad_start == expected_bad_start
        assert bad_stop == expected_bad_stop

    @pytest.mark.parametrize('missing', [
        slice(20032, 20033),  # First byte of header of frame 2.
        slice(20096, 20100),  # Part of payload of frame 2.
        slice(12000, 22000),  # Parts of 1-2.
        slice(30060, 30070),  # Part of header of frame 3.
        slice(40063, 40064)])  # Last byte of last frame.
    def test_bad_frames(self, missing, tmpdir):
        corrupted = (self.sample_bytes[:missing.start]
                     + self.sample_bytes[missing.stop:])
        bad_start, bad_stop = self.expected_bad_frames(missing)

        filename = str(tmpdir.join('corrupted.m5b'))
        with open(filename, 'wb') as fw:
            fw.write(corrupted)

        with mark5b.open(filename, 'rb', nchan=8, bps=2) as fr:
            for i in range(bad_start):
                header = fr.find_header(forward=True, maximum=40000,
                                        check=1)
                assert header['frame_nr'] == i
                fr.seek(16, 1)

            if bad_stop < 4:
                header = fr.find_header(forward=True, maximum=40000,
                                        check=1)
                assert header['frame_nr'] == bad_stop
                fr.seek(16, 1)

                for i in range(bad_stop+1, 4):
                    header = fr.find_header(forward=True, maximum=40000,
                                            check=1)
                    assert header['frame_nr'] == i
                    fr.seek(16, 1)
            else:
                with pytest.raises(HeaderNotFoundError):
                    fr.find_header(forward=True, maximum=40000,
                                   check=1)

    # Have to keep frame 0 intact, as well as header of frame 1.
    @pytest.mark.parametrize('missing', [
        slice(20032, 20033),  # First byte of header of frame 2.
        slice(20096, 20100),  # Part of payload of frame 2.
        slice(12000, 22000),  # Parts of 1-2.
        slice(30060, 30070),  # Part of header of frame 3.
        slice(40063, 40064)])  # Last byte of last frame.
    def test_missing_bytes(self, missing, tmpdir):
        corrupted = (self.sample_bytes[:missing.start]
                     + self.sample_bytes[missing.stop:])
        bad_start, bad_stop = self.expected_bad_frames(missing)

        filename = str(tmpdir.join('corrupted.m5b'))
        with open(filename, 'wb') as fw:
            fw.write(corrupted)
            # Add four more frames to ensure that _last_header is OK.
            for i in range(4, 8):
                header = self.frame3.header.copy()
                header.set_time(self.start_time + i/self.frame_rate,
                                frame_rate=self.frame_rate)
                header.update()
                frame = self.frame3.__class__(header, self.frame3.payload,
                                              valid=False)
                frame.tofile(fw)

        # Check that bad frames are found with verify only.
        with mark5b.open(filename, 'rs', nchan=8, bps=2,
                         kday=56000, verify=True) as fv:
            assert not fv.info.readable
            assert not fv.info.checks['continuous']
            assert 'continuous' in fv.info.errors
            expected_msg = 'While reading at {}'.format(
                bad_start * fv.samples_per_frame)
            assert expected_msg in fv.info.errors['continuous']

            # While only warnings are given when it is fixable.
            fv.verify = 'fix'
            assert fv.info.readable
            assert 'fixable' in fv.info.checks['continuous']
            assert 'continuous' in fv.info.warnings
            assert expected_msg in fv.info.warnings['continuous']
            assert 'problem loading frame' in fv.info.warnings['continuous']

        with mark5b.open(filename, 'rs', sample_rate=32*u.MHz,
                         kday=56000, nchan=8, bps=2) as fr:
            assert fr.start_time == self.start_time
            assert abs(fr.stop_time - self.stop_time
                       - 4 / self.frame_rate) < 1.*u.ns
            with pytest.warns(UserWarning,
                              match='problem loading frame'):
                data = fr.read()

        assert data.shape == (40000, 8)

        expected = self.data.copy().reshape(-1, 5000, 8)
        expected[bad_start:bad_stop] = 0.
        expected.shape = -1, 8
        expected = np.concatenate((expected, np.zeros_like(expected)))

        assert np.all(data == expected)


class TestCorruptFile:
    def setup(self):
        time = Time('2010-11-12T13:14:15')
        self.header0 = mark5b.Mark5BHeader.fromvalues(time=time)
        self.data = np.repeat([[-1, 1], [-3, 3]], 10000, axis=0)
        self.nchan = 2
        self.sample_rate = 100 * u.kHz
        self.kwargs = dict(nchan=self.nchan,
                           sample_rate=self.sample_rate,
                           ref_time=time)

    def fake_file(self, tmpdir, nframes=16):
        filename = str(tmpdir.join('fake.mark5b'))
        with mark5b.open(filename, 'ws', header0=self.header0,
                         sample_rate=self.sample_rate,
                         nchan=self.nchan) as fw:
            for _ in range(nframes):
                fw.write(self.data)
        return filename

    def test_fake_file(self, tmpdir):
        fake_file = self.fake_file(tmpdir)
        with mark5b.open(fake_file, 'rs', **self.kwargs) as fr:
            data = fr.read()

        data = data.reshape((-1,) + self.data.shape)
        assert np.all(data.astype(int) == self.data)

    def corrupt_copy(self, filename, missing):
        corrupt_name = filename.replace('.mark5b', '_corrupt.mark5b')
        with open(filename, 'rb') as fr, \
                open(corrupt_name, 'wb') as fw:
            fw.write(fr.read(missing.start))
            fr.seek(missing.stop)
            fw.write(fr.read())
        return corrupt_name

    @pytest.mark.parametrize('frame_nr', [1, 3, 5, slice(7, 10)])
    def test_missing_frames(self, frame_nr, tmpdir):
        if not isinstance(frame_nr, slice):
            frame_nr = slice(frame_nr, frame_nr+1)
        missing = slice(frame_nr.start * self.header0.frame_nbytes,
                        frame_nr.stop * self.header0.frame_nbytes)
        fake_file = self.fake_file(tmpdir)
        corrupt_file = self.corrupt_copy(fake_file, missing)
        with mark5b.open(corrupt_file, 'rs', **self.kwargs) as fr:
            with pytest.warns(UserWarning, match='problem loading'):
                data = fr.read()

        data = data.reshape((-1,) + self.data.shape)
        assert np.all(data[:frame_nr.start].astype(int) == self.data)
        assert np.all(data[frame_nr.stop:].astype(int) == self.data)
        assert np.all(data[frame_nr] == 0.)

    @pytest.mark.parametrize('missing_bytes', [
        slice(0, 10016),  # Remove whole last frame set.
        slice(0, 16),  # Remove header of last frame.
        slice(8, 16),  # Corrupt header of last frame.
        slice(0, 1),  # Corrupt header byte of last frame.
        slice(10, 11),  # Corrupt header byte of last frame.
        slice(15, 16),  # Corrupt header byte of last frame.
        slice(20, 21),  # Corrupt payload byte of last frame.
        slice(10015, 10016),  # last byte of last frame.
    ])
    def test_missing_end(self, missing_bytes, tmpdir):
        # In all these cases, the data read should just be short.
        missing = slice(missing_bytes.start + 15*self.header0.frame_nbytes,
                        missing_bytes.stop + 15*self.header0.frame_nbytes)
        fake_file = self.fake_file(tmpdir)
        corrupt_file = self.corrupt_copy(fake_file, missing)
        # If the sync pattern is still there, or the whole last frame
        # is missing, we'll get all but one.  Otherwise, the one-but-last
        # header will not be considered OK and we loose that frame too.
        if missing_bytes.start > 5 or (missing_bytes.start == 0
                                       and missing_bytes.stop == 10016):
            expected = 15
        else:
            expected = 14

        with mark5b.open(corrupt_file, 'rs', **self.kwargs) as fr:
            assert fr.size == expected * self.data.size
            data = fr.read()

        data = data.reshape((-1,) + self.data.shape)
        assert np.all(data.astype(int) == self.data)

    @pytest.mark.parametrize('missing_bytes,missing_frames', [
        (slice(10016, 20032), slice(1, 2)),  # Remove frame 1.
        (slice(20000, 20501), slice(1, 3)),  # Corrupt frames 1,2.
        (slice(20032, 20048), slice(1, 3)),  # Missing header 2.
    ])
    def test_missing_middle(self, missing_bytes, missing_frames, tmpdir):
        # In all these cases, some data will be missing.
        fake_file = self.fake_file(tmpdir)
        corrupt_file = self.corrupt_copy(fake_file, missing_bytes)
        with mark5b.open(corrupt_file, 'rs', **self.kwargs) as fr:
            assert fr.size == 16 * self.data.size
            with pytest.warns(UserWarning, match='problem loading frame'):
                data = fr.read()

        data = data.reshape((-1,) + self.data.shape)
        expected = np.stack([self.data] * 16)
        expected[missing_frames] = 0.
        assert np.all(data.astype(int) == expected)



# Licensed under the GPLv3 - see LICENSE
import pytest

from ..header import ParserDict, HeaderParserBase, HeaderParser


class ParserDictSetup:
    @staticmethod
    def create_parser(index, default):
        def parser(words):
            return words[index]
        return parser

    @staticmethod
    def get_default(index, default):
        return default

    @staticmethod
    def random_name(index, default):
        return index


class TestHeaderParserBase(ParserDictSetup):
    def setup_class(cls):
        cls.hp = {'0': (0, 'default0'),
                  '1': (1, 'default1')}

        class H(HeaderParserBase):
            parsers = ParserDict(cls.create_parser)
            indices = ParserDict(cls.random_name)
            defaults = ParserDict(cls.get_default)

        cls.H = H

    def test_parserdict(self):
        assert self.H.parsers.name == 'parsers'
        assert self.H.indices.name == 'indices'
        assert repr(self.H.parsers).startswith('ParserDict')
        assert 'Lazily evaluated' in self.H.defaults.__doc__

    def test_init(self):
        h = self.H(self.hp)
        words = ['first', 'second']
        assert h.parsers['0'](words) == 'first'
        assert h.defaults['1'] == 'default1'
        assert h.indices['1'] == 1


class TestHeaderParser:
    def setup_class(cls):
        cls.header_parser = HeaderParser(
            (('x0_16_4', (0, 16, 4)),
             ('x0_31_1', (0, 31, 1, False)),
             ('x1_0_32', (1, 0, 32)),
             ('x2_0_64', (2, 0, 64, 1 << 32))))

    def test_header_parser_update(self):
        extra = HeaderParser((('x4_0_32', (4, 0, 32)),))
        new = self.header_parser + extra
        assert len(new.keys()) == 5
        assert len(self.header_parser.keys()) == 4
        new = self.header_parser.copy()
        assert isinstance(new, HeaderParser)
        new.update(extra)
        assert len(new.keys()) == 5
        assert new['x4_0_32'] == (4, 0, 32)
        with pytest.raises(TypeError):
            self.header_parser + {'x4_0_32': (4, 0, 32)}
        with pytest.raises(ValueError):
            self.header_parser.copy().update(('x4_0_32', (4, 0, 32)))

    def test_header_parser_class(self):
        header_parser = self.header_parser.copy()
        words = [0x12345678, 0xffff0000, 0x0, 0xffffffff]
        header_parser['0_2_8'] = (0, 2, 8, 5)
        assert '0_2_8' in header_parser
        assert header_parser.defaults['0_2_8'] == 5
        assert header_parser.parsers['0_2_8'](words) == (words[0] >> 2) & 0xff
        # Check we can change and parsers will be reset.
        header_parser['0_2_8'] = (0, 1, 8, 3)
        assert '0_2_8' in header_parser
        assert header_parser.defaults['0_2_8'] == 3
        assert header_parser.parsers['0_2_8'](words) == (words[0] >> 1) & 0xff
        header_parser.update({'0_2_8': (0, 3, 8, 1)})
        assert '0_2_8' in header_parser
        assert header_parser.defaults['0_2_8'] == 1
        assert header_parser.parsers['0_2_8'](words) == (words[0] >> 3) & 0xff

        small_parser = HeaderParser((('0_2_8', (0, 2, 8, 4)),))
        header_parser2 = self.header_parser + small_parser
        assert header_parser2.parsers['0_2_8'](words) == (words[0] >> 2) & 0xff
        assert header_parser2.defaults['0_2_8'] == 4
        with pytest.raises(TypeError):
            header_parser + {'0_2_8': (0, 2, 8, 4)}
        with pytest.raises(Exception):
            self.HeaderParser((('0_2_32', (0, 2, 32, 4)),))
        with pytest.raises(Exception):
            self.HeaderParser((('0_2_64', (0, 2, 64, 4)),))



# Licensed under the GPLv3 - see LICENSE
"""
Definitions for VLBI Mark 4 payloads.

Implements a Mark4Payload class used to store payload words, and decode to
or encode from a data array.

For the specification, see
https://www.haystack.mit.edu/tech/vlbi/mark5/docs/230.3.pdf
"""
import operator

import numpy as np

from ..base.frame import FrameBase
from .header import Mark4Header
from .payload import Mark4Payload


__all__ = ['Mark4Frame']


class Mark4Frame(FrameBase):
    """Representation of a Mark 4 frame, consisting of a header and payload.

    Parameters
    ----------
    header : `~baseband.mark4.Mark4Header`
        Wrapper around the encoded header words, providing access to the
        header information.
    payload : `~baseband.mark4.Mark4Payload`
        Wrapper around the payload, provding mechanisms to decode it.
    valid : bool or None, optional
        Whether the data are valid.  If `None` (default), inferred from header.
        Note that ``header`` is updated in-place if `True` or `False`.
    verify : bool, optional
        Whether or not to do basic assertions that check the integrity
        (e.g., that channel information and number of tracks are consistent
        between header and data).  Default: `True`.

    Notes
    -----
    The Frame can also be read instantiated using class methods:

      fromfile : read header and payload from a filehandle

      fromdata : encode data as payload

    Of course, one can also do the opposite:

      tofile : method to write header and payload to filehandle

      data : property that yields full decoded payload

    One can decode part of the payload by indexing or slicing the frame.
    If the frame does not contain valid data, all values returned are set
    to ``self.fill_value``.

    A number of properties are defined: `shape`, `dtype` and `size` are
    the shape, type and number of complete samples of the data array, and
    `nbytes` the frame size in bytes.  Furthermore, the frame acts as a
    dictionary, with keys those of the header.  Any attribute that is not
    defined on the frame itself, such as ``.time`` will be looked up on the
    header as well.
    """

    _header_class = Mark4Header
    _payload_class = Mark4Payload

    def __init__(self, header, payload, valid=None, verify=True):
        self.header = header
        self.payload = payload
        if valid is not None:
            self.valid = valid
        if verify:
            self.verify()

    @property
    def valid(self):
        """Whether frame contains valid data.

        None of the error flags are set.
        """
        return not np.any(self.header['time_sync_error']
                          | self.header['internal_clock_error']
                          | self.header['processor_time_out_error']
                          | self.header['communication_error'])

    @valid.setter
    def valid(self, valid):
        if valid:
            self.header['time_sync_error'] = False
            self.header['internal_clock_error'] = False
            self.header['processor_time_out_error'] = False
            self.header['communication_error'] = False
        else:
            self.header['communication_error'] = True

    @classmethod
    def fromfile(cls, fh, ntrack, decade=None, ref_time=None, verify=True):
        """Read a frame from a filehandle.

        Parameters
        ----------
        fh : filehandle
            To read header from.
        ntrack : int
            Number of Mark 4 bitstreams.
        decade : int or None
            Decade in which the observations were taken.  Can instead pass an
            approximate ``ref_time``.
        ref_time : `~astropy.time.Time` or None
            Reference time within 4 years of the observation time.  Used only
            if ``decade`` is not given.
        verify : bool, optional
            Whether to do basic verification of integrity.  Default: `True`.
        """
        header = cls._header_class.fromfile(fh, ntrack, decade=decade,
                                            ref_time=ref_time, verify=verify)
        payload = cls._payload_class.fromfile(fh, header=header)
        return cls(header, payload, verify=verify)

    @classmethod
    def fromdata(cls, data, header=None, verify=True, **kwargs):
        """Construct frame from data and header.

        Parameters
        ----------
        data : `~numpy.ndarray`
            Array holding complex or real data to be encoded.  This should have
            the full size of a data frame, even though the part covered by the
            header will be ignored.
        header : `~baseband.mark4.Mark4Header` or None
            If not given, will attempt to generate one using the keywords.
        verify : bool, optional
            Whether to do basic checks of frame integrity (default: `True`).
        """
        if header is None:
            header = cls._header_class.fromvalues(verify=verify, **kwargs)
        assert data.shape[0] == header.samples_per_frame
        # Start of part not overwritten by header
        # (see calculation of header.samples_per_frame)
        start = header.nbytes * 8 // (header.ntrack // header.fanout)
        payload = cls._payload_class.fromdata(data[start:], header=header)

        return cls(header, payload, verify=verify)

    def __len__(self):
        """Number of samples (including those overwritten by header)."""
        return self.header.samples_per_frame

    def _get_payload_item(self, item):
        """Translate frame item to payload item, correcting for header part.

        For Mark 4 frames, part of the actual data is overwritten by the
        header.  In slicing a frame, these parts should be set to invalid.

        Parameters
        ----------
        item : int, slice, or tuple
            Sample indices.  An int represents a single sample, a slice
            a sample range, and a tuple of ints/slices a range for
            multi-channel data.

        Returns
        -------
        payload_item : int, slice, or None
            Part of the payload that should be decoded. `None` if all the
            requested data is in the invalid part.
        sample_index : tuple
            Any slicing beyond the sample number.
        data_shape : tuple
            Shape of the data array that the indexing will lead to.  Useful
            to create an array that partially needs to be set to invalid.
        ninvalid : int
            Number of points in the data that should be set to invalid (or that
            should be ignored for setting).

        Notes
        -----
        ``item`` is restricted to (tuples of) ints or slices, so one cannot
        access non-contiguous samples using advanced indexing.  If ``item``
        is a slice, a negative increment cannot be used.
        """
        nsample = len(self)
        valid_start = nsample - len(self.payload)
        if item == () or item == slice(None):
            # Short-cut for full payload.
            return slice(None), (), self.shape, valid_start

        if isinstance(item, tuple):
            sample_index = item[1:]
            item = item[0]
        else:
            sample_index = ()

        if isinstance(item, slice):
            start, stop, step = item.indices(nsample)

            data_shape = ((stop - start - 1) // step + 1,) + self.sample_shape
            payload_start = start - valid_start
            payload_stop = stop - valid_start
            if payload_start >= 0:
                # All requested data falls within the payload.
                payload_item = slice(payload_start, payload_stop, step)
                ninvalid = 0
            elif payload_stop > 0:
                # Some data overlaps with the payload.
                ninvalid, payload_start = divmod(payload_start, step)
                ninvalid = -ninvalid
                payload_item = slice(payload_start, payload_stop, step)
            else:
                # Everything is invalid.
                payload_item = None
                ninvalid = data_shape[0]
        else:
            # Not a slice. Maybe an index?
            try:
                item = operator.index(item)
            except Exception:
                raise TypeError("{0} object can only be indexed or sliced."
                                .format(type(self)))
            if item < 0:
                item += nsample

            if not (0 <= item < nsample):
                raise IndexError("{0} index out of range.".format(type(self)))

            data_shape = self.sample_shape
            payload_item = item - valid_start
            if payload_item >= 0:
                ninvalid = 0
            else:
                payload_item = None
                ninvalid = 1

        return payload_item, sample_index, data_shape, ninvalid

    def __getitem__(self, item=()):
        if isinstance(item, str):
            return self.header.__getitem__(item)

        # Normally, we would just pass on to the payload here, but for
        # Mark 4, we need to deal with data overwritten by the header.
        (payload_item, sample_index, data_shape,
         ninvalid) = self._get_payload_item(item)
        if not self.valid or payload_item is None:
            data = np.full(data_shape, self.fill_value, self.dtype)

        elif ninvalid == 0:
            data = self.payload[payload_item]

        else:
            # Note: Creating an empty array and setting part to invalid
            # is much faster than creating one pre-filled with invalid.
            data = np.empty(data_shape, self.dtype)
            data[:ninvalid] = self.fill_value
            data[ninvalid:] = self.payload[payload_item]

        if sample_index == ():
            return data
        else:
            return data[(Ellipsis,) + sample_index]

    def __setitem__(self, item, value):
        if isinstance(item, str):
            return self.header.__setitem__(item, value)

        # Normally, we would just pass on to the payload here, but for
        # Mark 4, we need to deal with data overwritten by the header.
        data = np.asanyarray(value)
        assert data.ndim <= 2

        (payload_item, sample_index, data_shape,
         ninvalid) = self._get_payload_item(item)

        if payload_item is None:
            return

        if ninvalid > 0:
            # See if data has enough dimensions so that we need to remove
            # the part that cannot set anything in the payload.
            if sample_index == ():
                sample_ndim = len(self.sample_shape)
            else:
                sample_ndim = np.empty(self.sample_shape)[sample_index].ndim
            if data.ndim == 1 + sample_ndim:
                data = data[ninvalid:]

        if sample_index != ():
            payload_item = (payload_item,) + sample_index

        self.payload[payload_item] = data

    data = property(__getitem__,
                    doc="Full decoded frame, with header part filled in.")



# Licensed under the GPLv3 - see LICENSE
"""Radio baseband I/O."""

# Affiliated packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *   # noqa
# ----------------------------------------------------------------------------

# Define minima for the documentation, but do not bother to explicitly check.
__minimum_python_version__ = '3.7'
__minimum_astropy_version__ = '4.0'

if not _ASTROPY_SETUP_:   # noqa
    # For egg_info test builds to pass, put package imports here.
    from .io import file_info, open  # noqa



# Licensed under the GPLv3 - see LICENSE
from operator import index
from math import gcd

import numpy as np
from astropy.utils import classproperty


__all__ = ['lcm', 'bcd_decode', 'bcd_encode', 'fixedvalue', 'byte_array',
           'CRC', 'CRCStack']


def lcm(a, b):
    """Calculate the least common multiple of a and b."""
    return abs(a * b) // gcd(a, b)


def bcd_decode(value):
    try:
        # Far faster than my routine for scalars
        return int('{:x}'.format(index(value)))
    except TypeError as exc:  # Might still be an array
        try:
            assert value.dtype.kind in 'iu'
        except Exception:
            raise exc

    d = np.arange(value.itemsize * 2)
    digits = (value[:, np.newaxis] >> d*4) & 0xf
    if digits.max() > 9:
        bad = value[(digits > 9).any(1)][0]
        raise ValueError("invalid BCD encoded value {0}={1}."
                         .format(bad, hex(bad)))
    return (digits * 10**d).sum(1)


def bcd_encode(value):
    try:
        # Far faster than my routine for scalars
        return int('{:d}'.format(index(value)), base=16)
    except TypeError as exc:  # Might still be an array
        try:
            assert value.dtype.kind in 'iu'
        except Exception:
            raise exc

    d = np.arange(value.itemsize * 2)
    digits = (value[:, np.newaxis] // 10**d) % 10
    return (digits << d*4).sum(1)


def byte_array(pattern):
    """Convert the pattern to a byte array.

    Parameters
    ----------
    pattern : ~numpy.ndarray, bytes, int, or iterable of int
        Pattern to convert.  If a `~numpy.ndarray` or `bytes` instance,
        a byte array view is taken.  If an (iterable of) int, the integers
        need to be unsigned 32 bit and will be interpreted as little-endian.

    Returns
    -------
    byte_array : `~numpy.ndarray` of byte
        With any elements of pattern stored in little-endian order.
    """
    if isinstance(pattern, (np.ndarray, bytes)):
        # Quick turn-around for input that is OK already:
        return np.atleast_1d(pattern).view('u1')

    pattern = np.array(pattern, ndmin=1)
    if (pattern.dtype.kind not in 'uif'
            or pattern.min() < 0
            or pattern.max() >= 1 << 32):
        raise ValueError('values have to fit in 32 bit unsigned int.')
    return pattern.astype('<u4').view('u1')


class fixedvalue(classproperty):
    """Property that is fixed for all instances of a class.

    Based on `astropy.utils.decorators.classproperty`, but with
    a setter that passes if the value is identical to the fixed
    value, and otherwise raises a `ValueError`.
    """
    def __set__(self, instance, value):
        fixed_value = self.__get__(instance, type(instance))
        if value != fixed_value:
            raise ValueError(
                f"'{self.fget.__name__}' can only be set to {fixed_value}.")


class CRC:
    """Cyclic Redundancy Check.

    See https://en.wikipedia.org/wiki/Cyclic_redundancy_check

    Once initialised, the instance can be used as a function that calculates
    the CRC, or one can use the ``check`` method to verify that the CRC in
    the lower bits of a value is correct.

    Parameters
    ----------
    polynomial : int
        Binary encoded CRC divisor. For instance, that used by Mark 5B headers
        is 0x18005, or x^16 + x^15 + x^2 + 1.

    See Also
    --------
    baseband.base.utils.CRCStack :
        for calculating CRC on arrays where each entry represents a bit.
    """

    def __init__(self, polynomial):
        self.polynomial = index(polynomial)

    def __len__(self):
        return self.polynomial.bit_length() - 1

    def __call__(self, stream):
        """Calculate CRC for the given stream.

        Parameters
        ----------
        stream : int or array of unsigned int
            The integer (or array of integers) to calculate the CRC for.

        Returns
        -------
        crc : int or array
            If an array, the crc will have the same dtype as the input stream.
        """
        return self._crc(stream, extend=True)

    def check(self, stream):
        """Check that the CRC at the end of athe stream is correct.

        Parameters
        ----------
        stream : int or array of unsigned int
            For an integer, the value is the stream to check the CRC for.
            For arrays, the dimension is treated as the index into the bits.
            A single stream would thus be of type `bool`. Unsigned integers
            represent multiple streams. E.g., for a 64-track Mark 4 header,
            the stream would be an array of ``np.uint64`` words.

        Returns
        -------
        ok : bool
             `True` if the calculated CRC is all zero (which should be the
             case if the CRC at the end of the stream is correct).
        """
        return self._crc(stream) == 0

    def _crc(self, stream, extend=False):
        try:
            scalar = index(stream)
        except TypeError:
            return self._crc_array(stream, extend=extend)
        else:
            return self._crc_scalar(scalar, extend=extend)

    def _crc_scalar(self, scalar, extend=False):
        """Internal function to calculate the CRC for a scalar."""
        # This routine uses bit_length() to find where the highest
        # remaining set bit is, thus skipping all the zeros where nothing
        # needs to be done.  It is about 15 times faster than a 1-element
        # array and can handle arbitrarily long integers.
        nbp = self.polynomial.bit_length()
        nbs = scalar.bit_length()
        while nbs >= nbp:
            scalar ^= self.polynomial << nbs-nbp
            nbs = scalar.bit_length()

        if extend:
            # Do the further iterations to *calculate* a CRC.
            return self._crc_scalar(scalar << nbp-1)

        return scalar

    def _crc_array(self, array, extend=False):
        """Internal function to calculate the CRC for an array."""
        # Here, the array contains individual entries for which the
        # CRC should be calculated.
        array = np.array(array, copy=True, dtype='u8')
        nbp = self.polynomial.bit_length()
        nbs = index(array.max()).bit_length()
        while nbs >= nbp:
            mask = (array & (1 << nbs-1)).astype(bool).astype(array.dtype)
            mask *= self.polynomial << nbs-nbp
            array ^= mask
            nbs = int(array.max()).bit_length()

        if extend:
            return self._crc_array(array << nbp-1)

        return array


class CRCStack(CRC):
    """Cyclic Redundancy Check for a bitstream.

    See https://en.wikipedia.org/wiki/Cyclic_redundancy_check

    Once initialised, the instance can be used as a function that calculates
    the CRC, or one can use the ``check`` method to verify that the CRC at
    the end of a stream is correct.

    This class is specifically for arrays in which multiple bit streams
    occupy different bit levels, and the dimension is treated as the
    index into the bits.  A single stream would thus be of type `bool`.
    Unsigned integers represent multiple streams. E.g., for a 64-track
    Mark 4 header, the stream would be an array of ``np.uint64`` words.

    Parameters
    ----------
    polynomial : int
        Binary encoded CRC divisor. For instance, that used by Mark 4 headers
        is 0x180f, or x^12 + x^11 + x^3 + x^2 + x + 1.

    See Also
    --------
    baseband.base.utils.CRC :
        for calculating CRC for a single value or an array of values.
    """
    def __init__(self, polynomial):
        super().__init__(polynomial)
        binary_str = '{:b}'.format(self.polynomial)
        self._npol_array = np.array([-int(bit) for bit in binary_str],
                                    dtype='i1')

    def check(self, stream):
        return np.all(self._crc(stream) == 0)

    def _crc(self, stream, extend=False):
        """Internal function to calculate the CRC for a bit stream."""
        ncrc = len(self)
        if extend:
            stream = np.hstack((stream, np.zeros(ncrc, stream.dtype)))
        else:
            stream = stream.copy()

        # Make an all-bits-one for each set item.
        pol_array = self._npol_array.astype(stream.dtype)
        for i, bits in enumerate(stream[:-ncrc]):
            stream[i:i+ncrc+1] ^= (bits & pol_array)

        return stream[-ncrc:]
